# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

from diffusers.configuration_utils import ConfigMixin as HF_ConfigMixin

from ..decorators import Register

logger = logging.getLogger(__name__)


class ConfigMixin(HF_ConfigMixin):
    pass


def extract_init_dict(config_model):
    return config_model.extract_init_dict(config_model._internal_dict)


def auto_cls_from_pretrained(
    register: Register, mixin_class: type, path: str, **kwargs
):

    if not hasattr(mixin_class, 'load_config'):
        raise RuntimeError(f'Class `{mixin_class}` do not has attribute `load_config`.')

    config = mixin_class.load_config(path)
    if isinstance(config, tuple):
        config = config[0]

    cls_name = config.get('_class_name', None)
    if cls_name is None:
        raise RuntimeError('`_class_name` is not contained in config.')

    try:
        cls = register[cls_name]
    except:
        raise RuntimeError(f'_class_name `{cls_name}` has not been registered.')

    logging.debug(f'Create instance for `{cls_name}`')

    return cls.from_pretrained(path, **kwargs)



import os
import time
import logging
from huggingface_hub import snapshot_download


def retry(func, *args, max_retries=3, delay=0.5, silent=False, **kwargs):
    """
    Retry function if it raises an exception.

    Args:
        func (function): The function to be retried.
        *args (tuple): Positional arguments for the function.
        max_retries (int): Maximum number of retries. Default is 3.
        delay (float): Delay between retries in seconds. Default is 0.5.
        **kwargs (dict): Keyword arguments for the function.

    Returns:
        The result of the function.
    """
    for i in range(max_retries):
        try:
            result = func(*args, **kwargs)
            if not silent:
                logging.info(
                    f"Function {func.__name__} executed successfully on attempt {i+1}"
                )
            return result
        except Exception as e:
            if not silent:
                logging.error(f"Function {func.__name__} failed on attempt {i+1}: {e}")
            if i == max_retries - 1:
                raise e
            time.sleep(delay)


def hf_snapshot_download(
    repo_id, repo_type="model", max_retries=3, mirror=None, **kwargs
):
    origin_hf_endpoint = (
        None if 'HF_ENDPOINT' not in os.environ else os.environ['HF_ENDPOINT']
    )
    if mirror is not None:
        os.environ['HF_ENDPOINT'] = mirror
    try:
        func = lambda: snapshot_download(repo_id=repo_id, repo_type=repo_type, **kwargs)
        return retry(func, max_retries=max_retries)
    finally:
        if origin_hf_endpoint is not None:
            os.environ['HF_ENDPOINT'] = origin_hf_endpoint



from collections import deque, Counter
from typing import Iterable, Set, Any, Literal


class IntervalCounter:
    def __init__(self, interval: int, couter_type: Literal['start', 'end'] = 'start'):
        self.interval = interval
        self.count = 0
        self.couter_type = couter_type

    def step(self):
        # self.count += 1
        # return self.is_interval()
        if self.couter_type == 'start':
            result = self.is_interval()
            self.count += 1
        else:
            self.count += 1
            result = self.is_interval()
        return result

    def reset(self):
        self.count = 0

    def is_interval(self):
        return self.count % self.interval == 0


class RecentKCommon:

    def __init__(self, k: int):
        if k < 1:
            raise ValueError("k must be >= 1")
        self.k = k
        self.window: deque[Set[Any]] = deque()
        self.counts: Counter = Counter()

    def push(self, lst: Iterable[Any]) -> Set[Any]:
        s = set(lst)
        self.window.append(s)
        for e in s:
            self.counts[e] += 1

        if len(self.window) > self.k:
            old = self.window.popleft()
            for e in old:
                self.counts[e] -= 1
                if self.counts[e] == 0:
                    del self.counts[e]
        return self.current_common()

    def current_common(self) -> Set[Any]:
        return {e for e, c in self.counts.items() if c == self.k}



# Copyright 2024 Sijin Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from torch.utils.data import Dataset


class DatasetWrapper(Dataset):
    def __init__(self, dataset: Dataset) -> None:
        super().__init__()
        self.dataset = dataset

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            if hasattr(self.dataset, name):
                return getattr(self.dataset, name)
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )

    def __hasattr__(self, name):
        return hasattr(self.dataset, name) or super().__hasattr__(name)


class DownSampleDatasetWrapper(DatasetWrapper):
    def __init__(self, dataset: Dataset, ratio: float = 0.1):
        super().__init__(dataset)
        self.ratio = ratio
        self.length = round(len(self.dataset) * self.ratio)

    def __getitem__(self, index: int):
        original_index = round(index / self.ratio)
        original_index = min(original_index, len(self.dataset) - 1)
        data = self.dataset[original_index]
        return original_index, data

    def __len__(self):
        return self.length


class RandomDownSampleDatasetWrapper(DatasetWrapper):
    def __init__(self, dataset: Dataset, ratio: float = 0.1, seed: int = 42):
        super().__init__(dataset)
        self.ratio = ratio
        self.length = round(len(self.dataset) * self.ratio)
        self.indices = np.random.RandomState(seed).choice(
            len(self.dataset), self.length, replace=False
        )

    def __getitem__(self, index: int):
        original_index = self.indices[index]
        data = self.dataset[original_index]
        return original_index, data

    def __len__(self):
        return self.length


class DatasetIndexWrapper(DatasetWrapper):
    def __init__(self, dataset: Dataset):
        super().__init__(dataset)

    def __getitem__(self, index: int):
        data = self.dataset[index]
        return index, data

    def __len__(self):
        return len(self.dataset)



# Copyright 2024 Hongyao Yu and Sijin Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import Literal
from collections import defaultdict, OrderedDict

try:
    import torch

    _has_torch = True
except:
    _has_torch = False


class ListAccumulator:
    """For accumulating sums over `n` variables."""

    def __init__(self, n: int):
        self.data = [0] * n
        self.num = 0

    def add(self, *args, add_num: int = 1, add_type: Literal['mean', 'sum'] = 'mean'):
        """adding data to the data list"""
        assert len(args) == len(self.data)
        mul_coef = add_num if add_type == 'mean' else 1
        self.num += add_num
        for i, add_item in enumerate(args):
            if _has_torch and isinstance(add_item, torch.Tensor):
                add_item = add_item.item()
            self.data[i] += add_item * mul_coef

    def reset(self):
        """reset all data to 0"""
        self.data = [0] * len(self.data)
        self.num = 0

    def __getitem__(self, idx):
        return self.data[idx]

    def __len__(self):
        return len(self.data)

    def avg(self, idx=None):
        """Calculate average of the data specified by `idx`. If idx is None, it will calculate average of all data.

        Args:
            idx (int, optional): subscript for the data list. Defaults to None.

        Returns:
            int | list: list if idx is None else int
        """
        num = 1 if self.num == 0 else self.num
        if idx is None:
            return [d / num for d in self.data]
        else:
            return self.data[idx] / num


class DictAccumulator:
    def __init__(self) -> None:
        self.data = OrderedDict()  # defaultdict(lambda : 0)
        self.num = 0

    def reset(self):
        """reset all data to 0"""
        self.data = OrderedDict()  # defaultdict(lambda : 0)
        self.num = 0

    def add(
        self, add_dic: OrderedDict, add_num=1, add_type: Literal['mean', 'sum'] = 'mean'
    ):
        mul_coef = add_num if add_type == 'mean' else 1
        self.num += add_num
        for key, val in add_dic.items():
            if _has_torch and isinstance(val, torch.Tensor):
                val = val.item()
            if key not in self.data.keys():
                self.data[key] = 0
            self.data[key] += val * mul_coef

    def __getitem__(self, key):
        return self.data[key]

    def __len__(self):
        return len(self.data)

    def avg(self, key=None):
        num = 1 if self.num == 0 else self.num
        if key is None:
            res = copy.deepcopy(self.data)
            for k in self.data:
                res[k] /= num
            return res
        else:
            return self.data[key] / num



import functools
from typing import *

_T = TypeVar('_T')
_TC = TypeVar('_TC')
_P = ParamSpec('_P')


def convert_to_collection(
    collection_builder: Callable[_P, _TC], convert_none: bool = False
):
    """A decorator to convert the result of the function.

    Args:
        collection_builder (Callable[_P, _TC]): The constructor of the collection class.
        convert_None (bool, optional): If the function returns `None`, it will return an empty collection if `convert_none` is True, otherwise return `None`. Defaults to False.
    """

    def _convert(func: Callable[_P, Optional[Iterable[_T]]]) -> Callable[_P, _TC]:

        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            result: Optional[Iterable[_T]] = func(*args, **kwargs)

            if result is None:
                if convert_none:
                    return collection_builder([])
                else:
                    return None
            else:
                return collection_builder(result)

        return _wrapper

    return _convert



import torch
import torch.version

print(torch.version.cuda)

import torch
import my_cuda_add

# 使用 CUDA 张量
x = torch.randn(1024, device="cuda")
y = torch.randn(1024, device="cuda")
out = torch.empty_like(x)

# 调用扩展函数
my_cuda_add.add(x, y, out)

print("x + y = ", out)

import latexify



import os
import logging

# from .configuration.autosave_utils import auto_register_save_load
from .auto import auto_register_save_load
from .welcome import add_welcome_msg, welcome_print
from .constant import USER_NAME
from .environ import _environ_init

logger = logging.getLogger(__name__)


def _register_save_load():

    logger.debug("Registering save and load functions for phoenixcat.")

    # if is_accelerate_available():
    #     from accelerate import Accelerator, DataLoaderConfiguration

    #     auto_register_save_load(DataLoaderConfiguration)

    #     auto_register_save_load(
    #         Accelerator,
    #         ignore_list=[
    #             'dispatch_batches',
    #             'split_batches',
    #             'even_batches',
    #             'use_seedable_sampler',
    #         ],
    #     )


def _welcome():
    welcome_print(USER_NAME)


def _init_fn():
    _environ_init()
    _register_save_load()
    # _welcome()



import json
import logging
import os
import requests
import time

logger = logging.getLogger(__name__)


def get_json(
    url: str,
    local_file: os.PathLike = None,
    mkdirs: bool = True,
    retry: int = 10,
    wait: float = 2.0
):
    """Fetch JSON data from a given URL.

    Args:
        url (str): The URL from which to download the JSON data.
        local_file (os.PathLike, optional): Path to the file where the downloaded JSON data should be saved.
            If None, the data is not saved to a file. Defaults to None.
        mkdirs (bool, optional): If True, creates the directory path for the local_file if it does not already exist.
            Defaults to True.
        retry (int, optional): Number of retry attempts to download the data in case of failure. Defaults to 10.
        wait (float, optional): Time in seconds to wait between retry attempts. Defaults to 2.0 seconds.

    Returns:
        dict: A dictionary containing the JSON data if the download was successful. None if not successful.
    """
    for i in range(retry):
        try:
            response = requests.get(url)
        except Exception as e:
            response = None
            logger.warning(f"Failed to download from {url}: {str(e)}, retry {i + 1} / {retry} after {wait} sec.")
            time.sleep(wait)
    
    if response is not None:
        data = json.loads(response.text)
        logger.info(f"Get json data from {url}.")
        if local_file is not None:
            dirname = os.path.dirname(local_file)
            if mkdirs and dirname != "":
                os.makedirs(dirname, exist_ok=True)
            with open(local_file, "w") as file:
                json.dump(data, file, indent=4)
            logger.debug(f"Save {local_file}.")
        return data
    else:
        logger.error(f"Failed to get data from {url}.")
    


import os
import hashlib
import pickle

from .constant import ASSETS_CACHE

WELCOME_ASSERT_PATH = os.path.join(ASSETS_CACHE, "welcome.pkl")


def _encode(name):
    return hashlib.md5(name.encode()).hexdigest()


def add_welcome_msg(name: str, welcome_msg: str):

    if not isinstance(welcome_msg, str):
        raise RuntimeError('The `welcome_msg` should be str.')

    if not os.path.exists(WELCOME_ASSERT_PATH):
        data = {}
    else:
        with open(WELCOME_ASSERT_PATH, 'rb') as f:
            data = pickle.load(f)

    encode_name = _encode(name)
    data[encode_name] = welcome_msg

    with open(WELCOME_ASSERT_PATH, 'wb') as f:
        pickle.dump(data, f)


def _default_welcome(name):
    print(f'Hello {name}!')


def welcome_print(name: str):

    if not os.path.exists(WELCOME_ASSERT_PATH):
        _default_welcome(name)
        return

    encode_name = _encode(name)
    with open(WELCOME_ASSERT_PATH, 'rb') as f:
        data = pickle.load(f)

    welcome_msg = data.get(encode_name, None)

    if welcome_msg is None:
        _default_welcome(name)
        return

    for msg in welcome_msg.split('\n'):
        print(msg)



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import shutil
import inspect
from pathlib import Path
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Dict, Any, Union, Literal
from contextlib import contextmanager
from functools import wraps

from .load import load_json, load_yaml, load_torchobj, load_pickle
from .save import (
    safe_save_as_json,
    safe_save_as_yaml,
    safe_save_torchobj,
    safe_save_as_pickle,
)
from ..auto.autosave_utils import is_json_serializable


class CacheManager:

    cache_info_filename = 'cache.json'
    files_dirname = 'files'

    def __init__(self, root_dir: str):
        self.root_dir = Path(root_dir)
        # self.root_dir.mkdir(exist_ok=True)
        os.makedirs(self.root_dir, exist_ok=True)
        self.cache_info_file = self.root_dir / self.cache_info_filename

        self._cache_info = None

    @property
    def cache_info(self):
        if self._cache_info is None:
            if not self.cache_info_file.exists():
                with open(self.cache_info_file, 'r') as f:
                    self._cache_info = json.load(f)
            else:
                self._cache_info = {}
        return self._cache_info

    def get_cache_dir(self, name):
        cnt = len(self.cache_info)
        if name in self.cache_info:
            return self.cache_info[name]
        else:
            cache_dir = self.root_dir / self.files_dirname / f'{cnt}'
            # cache_dir.mkdir(exist_ok=True)
            os.makedirs(cache_dir, exist_ok=True)
            self.cache_info[name] = cache_dir
            self.dump_cache_info()
            return cache_dir

    def exists(self, name):
        return name in self.cache_info

    def dump_cache_info(self):
        with open(self.cache_info_file, 'w') as f:
            json.dump(self.cache_info, f, indent=2)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dump_cache_info()


class FolderManager:

    def __init__(self, root, read_only=True):
        self.root = Path(root)
        self.read_only = read_only

        if not self.read_only:
            # self.root.mkdir(exist_ok=True)
            os.makedirs(self.root, exist_ok=True)

        self.ptr = self.root.resolve()

    def get_target_path(self, path: str = None, root=False):
        ptr = self.root if root else self.ptr
        if path is not None:
            ptr = ptr / path
        return ptr.resolve()

    def get_related_to_root(self):
        return self.ptr.relative_to(self.root)

    def cd(self, path: str, root=False):
        self.ptr = self.get_target_path(path, root=root)
        return self.ptr

    def ls(self, path: str = None):
        ptr = self.get_target_path(path)
        return os.listdir(str(ptr))

    def makedirs(self, path: str = None):
        if self.read_only:
            raise PermissionError('Read only mode')
        ptr = self.get_target_path(path)
        # ptr.mkdir(exist_ok=True)
        os.makedirs(ptr, exist_ok=True)
        return ptr

    def rm(self, path: str = None):
        if self.read_only:
            raise PermissionError('Read only mode')
        ptr = self.get_target_path(path)
        shutil.rmtree(ptr)

    def open(self, path: str = None, mode: str = 'r', open_kwargs=None, root=False):
        is_read = mode.startswith('r')

        if self.read_only and not is_read:
            raise PermissionError('Read only mode')

        ptr = self.get_target_path(path, root=root)
        if is_read and not ptr.exists():
            raise FileNotFoundError(f'{ptr} not found')

        if ptr.is_dir():
            raise IsADirectoryError(f'{ptr} is a directory')

        # ptr.parent.mkdir(exist_ok=True)
        os.makedirs(ptr.parent, exist_ok=True)

        if open_kwargs is None:
            open_kwargs = {}
        return ptr.open(mode, **open_kwargs)

    def pwd(self, related=False):
        if related:
            return self.get_related_to_root()
        else:
            return self.ptr

    def is_file(self, path: str = None):
        return self.get_target_path(path).is_file()

    def is_dir(self, path: str = None):
        return self.get_target_path(path).is_dir()

    def parent(self):
        self.ptr = self.ptr.parent
        return self.ptr

    def __str__(self):
        return str(self.ptr)


class DualFolderManager:

    def __init__(self, read_root, write_root):
        self.read_manager = FolderManager(read_root, read_only=True)
        self.write_manager = FolderManager(write_root, read_only=False)

    def get_target_path(self, path: str = None, root=False):
        return (
            self.read_manager.get_target_path(path, root=root),
            self.write_manager.get_target_path(path, root=root),
        )

    def cd(self, path: str, root=False):
        self.read_manager.cd(path, root=root)
        self.write_manager.cd(path, root=root)

    def ls(self, path: str = None):
        return self.read_manager.ls(path)

    def is_file(self, path: str = None):
        return self.read_manager.is_file(path)

    def is_dir(self, path: str = None):
        return self.read_manager.is_dir(path)

    @contextmanager
    def open(
        self,
        path: str = None,
        read_mode: str = 'r',
        write_mode: str = 'w',
        open_kwargs=None,
        root: bool = False,
        write_extension: str = None,
    ):
        read_file = self.read_manager.open(
            path, mode=read_mode, open_kwargs=open_kwargs, root=root
        )

        if write_extension is not None:
            if not write_extension.startswith('.'):
                write_extension = '.' + write_extension
            path = str(path).rsplit('.', 1)[0] + write_extension
        write_file = self.write_manager.open(
            path, mode=write_mode, open_kwargs=open_kwargs, root=root
        )
        try:
            yield read_file, write_file
        finally:
            read_file.close()
            write_file.close()

    def copy(self, path: str = None, root=False):
        src_file = self.read_manager.get_target_path(path, root=root)
        dst_file = self.write_manager.get_target_path(path, root=root)
        # dst_file.parent.mkdir(exist_ok=True)
        os.makedirs(dst_file.parent, exist_ok=True)
        if src_file.is_dir():
            shutil.copytree(src_file, dst_file)
        else:
            shutil.copyfile(src_file, dst_file)

        return src_file, dst_file

    def pwd(self, related=False):
        return self.read_manager.pwd(related=related)

    def parent(self):
        self.read_manager.parent()
        self.write_manager.parent()


class RecordManager:

    tag: str = ".cache.phoenixcat"
    _format2suffix = {"torch": ".pt", "pickle": ".pkl"}
    file: str = None

    anchor_name = "_params"
    result_name = "_results"

    def __init__(
        self,
        file: str = None,
        file_format: Literal['json', 'yaml'] = 'json',
        non_serializable_format: Literal["torch", "pickle"] = "pickle",
    ):
        file = str(file)
        self.file = file or self.file

        self.file_format = file_format
        self.non_serializable_format = non_serializable_format

        self._is_init = False

    def _delay_init(self):

        if self.file is None:
            raise FileNotFoundError(f"File {self.file} not found.")

        self.load_fn = load_json if self.file_format == 'json' else load_yaml
        self.save_fn = (
            safe_save_as_json if self.file_format == 'json' else safe_save_as_yaml
        )
        self.non_serializable_load_fn = (
            load_torchobj if self.non_serializable_format == 'torch' else load_pickle
        )
        self.non_serializable_save_fn = (
            safe_save_torchobj
            if self.non_serializable_format == 'torch'
            else safe_save_as_pickle
        )

        if not os.path.exists(self.file):
            self.cache_dir = self._get_init_cache_dir(self.file)
            self.data = {"_cache_dir": self.cache_dir}
        else:
            self.data = self.load_fn(self.file)
            self.cache_dir = self.data.get("_cache_dir")

        os.makedirs(self.cache_dir, exist_ok=True)
        self.cache_info_file = os.path.join(
            self.cache_dir, f"cache_info.{self.file_format}"
        )
        if os.path.exists(self.cache_info_file):
            self.cache_info = self.load_fn(self.cache_info_file)
        else:
            self.cache_info = {}

        self._is_init = True

    # def set

    def _get_init_cache_dir(self, original_file):
        basename = os.path.basename(original_file)
        if '.' in basename:
            basename = basename.rsplit('.', 1)[0]
        return os.path.join(
            os.path.dirname(original_file),
            self.tag,
            basename,
        )

    def get_dict_value(self, data: Dict, key: str, anchor_values=None):

        if key in self.cache_info:
            return True, self.non_serializable_load_fn(self.cache_info[key])

        keys = key.split(".")
        for k in keys[:-1]:
            if k not in data:
                return False, None
            data = data[k]

        last_key = keys[-1]
        if last_key not in data:
            return False, None
        # return True, data[keys[-1]]

        result = data[keys[-1]]
        if anchor_values is None:
            return True, result

        if not isinstance(result, list):
            raise ValueError(f"Anchor values are not supported for non-list values.")

        for value in result:
            if value.get(self.anchor_name, None) == anchor_values:
                return True, value

        return False, None

    def set_dict_value(self, data: Dict, key: str, value, anchor_values=None):
        json_serializable = is_json_serializable(value)
        if not json_serializable:
            folder_values = [self.cache_dir, *key.split(".")]
            save_name = (
                folder_values[-1] + self._format2suffix[self.non_serializable_format]
            )
            folder_values = folder_values[:-1]
            folder = os.path.join(*folder_values)
            os.makedirs(folder, exist_ok=True)
            save_path = os.path.join(folder, save_name)
            self.cache_info[key] = save_path
            self.non_serializable_save_fn(value, save_path)

        keys = key.split(".")
        for k in keys[:-1]:
            if k not in data:
                data[k] = {}
            data = data[k]

        final_key = keys[-1]
        if anchor_values is None:
            data[final_key] = value if json_serializable else f"@{save_path}"
        else:
            if final_key not in data:
                data[final_key] = []
            data[final_key].append(
                {
                    self.anchor_name: anchor_values,
                    self.result_name: value if json_serializable else f"@{save_path}",
                }
            )

    def save(self):
        file = self.file
        self.save_fn(self.data, file)
        self.save_fn(self.cache_info, self.cache_info_file)

    def cache_apply(self, key, anchor=None, key_anchor=None, func=None):
        wrapper = self.__call__(key=key, anchor=anchor, key_anchor=key_anchor)
        if func is None:
            return wrapper
        return wrapper(func)

    def __call__(
        self,
        key,
        anchor: None | str | List[str] = None,
        key_anchor: None | str | List[str] = None,
    ):

        if not self._is_init:
            self._delay_init()

        if key_anchor is None:
            key_anchor = []
        elif isinstance(key_anchor, str):
            key_anchor = [key_anchor]

        def _inner_func(func):
            @wraps(func)
            def wrapper(*args, **kwargs):

                nonlocal anchor

                if anchor is None:
                    anchor_values = None
                else:
                    if isinstance(anchor, str):
                        anchor = [anchor]

                    init_kwargs = {
                        k: v for k, v in kwargs.items() if not k.startswith("_")
                    }

                    signature = inspect.signature(func)
                    parameters = {
                        name: p.default for (name, p) in (signature.parameters.items())
                    }
                    for arg, name in zip(args, parameters.keys()):
                        parameters[name] = arg
                    for k, v in init_kwargs.items():
                        parameters[k] = v
                    parameters.pop("self", None)

                    anchor_values = {k: parameters.get(k, None) for k in anchor}
                    key_anchor_values = [
                        str(parameters.get(k, "_default")) for k in key_anchor
                    ]

                    _key = '.'.join([key] + key_anchor_values)

                    if not is_json_serializable(anchor_values):
                        print(args)
                        print(kwargs)
                        print(anchor_values)
                        raise ValueError("anchor values must be json serializable, ")

                has_cache, value = self.get_dict_value(
                    self.data, _key, anchor_values=anchor_values
                )
                if has_cache:
                    return value
                return_value = func(*args, **kwargs)
                self.set_dict_value(
                    self.data, _key, return_value, anchor_values=anchor_values
                )
                self.save()
                return return_value

            return wrapper

        return _inner_func


record_manager: RecordManager = RecordManager()


def set_record_manager_path(
    file: str,
    file_format: Literal['json', 'yaml'] = 'json',
    non_serializable_format: Literal["torch", "pickle"] = "pickle",
):
    global record_manager
    record_manager.file = file
    record_manager.file_format = file_format
    record_manager.non_serializable_format = non_serializable_format
    record_manager._is_init = False



from .data.dataloader import DataLoaderConfig, getDataLoader
from .trainer_utils import (
    TrainerMixin,
    register_to_run_one_epoch,
    register_to_run_one_iteration,
)
from .losses import TorchLoss
from .optimization import (
    OptimizationConfig,
    OptimizationManager,
    SingleOptimizationManager,
    get_optimizer,
    get_lr_scheduler,
)
from .train_pipeline import TrainPipelineMixin



# Copyright 2025 Hongyao Yu and Sijin Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
from typing import get_args, Optional
from types import UnionType
from dataclasses import is_dataclass, dataclass
from itertools import chain

from ..files.save import safe_save_as_json


def dict2dataclass(_data, _class):
    if isinstance(_data, dict):
        fieldtypes = {f.name: f.type for f in _class.__dataclass_fields__.values()}
        kwargs = {}
        for f in fieldtypes:
            value = _data.get(f)
            fieldtype = fieldtypes[f]
            if value is None:
                kwargs[f] = None
            else:
                _success = False

                searched_types = chain([fieldtype], get_args(fieldtype))

                for elem_type in searched_types:
                    try:
                        kwargs[f] = dict2dataclass(value, elem_type)
                        _success = True
                        break
                    except:
                        pass

                if not _success:
                    raise TypeError(f"Cannot convert {value} to {fieldtype}")

        return _class(**kwargs)

    elif isinstance(_data, list):
        if hasattr(_class, '__origin__') and _class.__origin__ == list:
            elem_type = get_args(_class)[0]
            return [dict2dataclass(d, elem_type) for d in _data]
        else:
            raise TypeError("Expected a list type annotation.")
    else:
        return _data


# _auto_save_configs_name = "_auto_save_configs"


def config_dataclass_wrapper(config_name='config.json'):

    def _inner_wrapper(cls):

        @classmethod
        def from_config(cls, config_or_path: dict | str):
            if not isinstance(config_or_path, dict):
                config_or_path: str

                if not config_or_path.endswith(config_name):
                    config_or_path = os.path.join(config_or_path, config_name)

                with open(config_or_path, 'r') as f:
                    config_or_path = json.load(f)

            # config = config_or_path
            # return cls(**config)
            return dict2dataclass(config_or_path, cls)

        def get_config(self):
            config = {}
            for k, v in self.__dict__.items():
                if is_dataclass(v):
                    config[k] = v.get_config()
                elif isinstance(v, list):
                    config[k] = [i.get_config() if is_dataclass(i) else i for i in v]
                else:
                    config[k] = v
            return config

        def save_config(self, path: str):
            os.makedirs(path, exist_ok=True)
            path = os.path.join(path, config_name)
            # safe_save_as_json(self.__dict__, path)
            save_info = self.get_config()

            safe_save_as_json(save_info, path)

        cls.from_config = cls.from_pretrained = from_config
        cls.save_config = cls.save_pretrained = save_config
        cls.get_config = get_config

        return cls

    return _inner_wrapper



from enum import Enum
from typing import Literal


class Color(Enum):
    black = 30
    red = 31
    green = 32
    yellow = 33
    blue = 34
    purple = 35
    cyan = 36
    white = 37

    def __str__(self):
        return str(self.value)


class BackgroundColor(Enum):
    black = 40
    red = 41
    green = 42
    yellow = 43
    blue = 44
    purple = 45
    cyan = 46
    white = 47

    def __str__(self):
        return str(self.value)


class FontStyle(Enum):
    normal = 0
    bold = 1
    underline = 4
    blink = 5
    reverse = 7
    hidden = 8

    def __str__(self):
        return str(self.value)


def get_enum_value(value: int | str | None, enum_cls):

    if value is None:
        return None

    if isinstance(value, (int, enum_cls)):
        return str(value)

    return str(enum_cls[value].value)


def get_colored_str(
    content: str,
    color: (
        int
        | Literal['black', 'red', 'green', 'yellow', 'blue', 'purple', 'cyan', 'white']
        | None
    ) = None,
    background: (
        int
        | Literal['black', 'red', 'green', 'yellow', 'blue', 'purple', 'cyan', 'white']
        | None
    ) = None,
    style: (
        int
        | Literal['normal', 'bold', 'underline', 'blink', 'reverse', 'hidden']
        | None
    ) = None,
) -> str:
    # "\033[1;35;42m高亮紫色文字绿色背景\033[0m"

    font_styles = []
    for value, enum_cls in [
        (color, Color),
        (background, BackgroundColor),
        (style, FontStyle),
    ]:
        if value is not None:
            font_styles.append(get_enum_value(value, enum_cls))
    if not font_styles:
        return content
    return f"\033[{';'.join(font_styles)}m{content}\033[0m"



import importlib
from typing import Sequence, Union

import logging

logger = logging.getLogger(__name__)


def get_obj_from_str(
    strings: Union[str, Sequence[str]], reload=False, raise_exception=False
):
    if not isinstance(strings, str):
        logger.debug(
            f"Recieve `strings` to be {strings} (type = {type(strings)}), not str."
        )
        strings = list(strings)
        strings = ".".join(strings)
    module, cls = strings.rsplit(".", 1)
    if reload:
        _module = importlib.import_module(module)
        importlib.reload(_module)
        logger.debug(f"Reload module `{module}`")

    module = importlib.import_module(module)
    if not hasattr(module, cls):
        if raise_exception:
            raise AttributeError(f"Module `{module}` has no attribute `{cls}`")
        return None
    return getattr(module, cls)


def get_attribute_from_obj(obj, attribute: str):
    for att in attribute.split("."):
        obj = getattr(obj, att)
    return obj



import os
import logging
from typing import Literal


def init_logger(
    file_name: os.PathLike = None,
    console_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "WARNING",
    file_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "DEBUG",
    msgfmt: str = "%(asctime)s [%(levelname)s] %(name)s: %(message)s",
    datefmt: str = "%Y-%m-%d %H:%M:%S",
):
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(msgfmt, datefmt=datefmt)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(getattr(logging, console_level))
    console_handler.setFormatter(formatter)

    dirname = os.path.dirname(file_name)
    if dirname != "":
        os.makedirs(dirname, exist_ok=True)
    file_handler = logging.FileHandler(file_name)
    file_handler.setLevel(file_level)
    file_handler.setFormatter(formatter)

    logger.addHandler(console_handler)
    logger.addHandler(file_handler)



import logging

import torch
import numpy as np
from diffusers.configuration_utils import register_to_config

from ..trainer_utils import TrainerMixin
from .base import CallbackMixin, register_callback

logger = logging.getLogger(__name__)


class AvgDataCallback(CallbackMixin):

    def __init__(self, fullname: str):
        super().__init__()
        self.fullname = fullname
        self.fullnames = fullname.split('.')
        self.name = self.fullnames[-1]
        self.data = 0
        self.cnt = 0

    def epoch_begin(self, trainer: TrainerMixin):
        self.data = 0
        self.cnt = 0

    def epoch_end(self, trainer: TrainerMixin):
        avg_data = 0 if self.cnt == 0 else self.data / self.cnt
        logger.info(f'{self.name}: {avg_data:.6f}')

    def step_end(self, trainer: TrainerMixin):
        data = trainer
        for n in self.fullnames:
            if not hasattr(data, n):
                raise RuntimeError(f'`trainer` do not has attribute `{self.fullname}`')
            data = getattr(data, n)

        self.cnt += 1
        self.data += float(torch.mean(data))


@register_callback
class AvgLossCallback(AvgDataCallback):

    @register_to_config
    def __init__(self):
        super().__init__('train_temp_values.loss')



import time
import logging

from diffusers.configuration_utils import register_to_config

from .base import CallbackMixin, register_callback
from ..trainer_utils import TrainerMixin
from ...format import format_time_invterval

logger = logging.getLogger(__name__)


@register_callback
class TimeCallback(CallbackMixin):

    @register_to_config
    def __init__(self):
        super().__init__()

    def epoch_begin(self, trainer: TrainerMixin):
        self.t = time.time()

    def epoch_end(self, trainer: TrainerMixin):
        t = time.time() - self.t
        t_format = format_time_invterval(t)
        logger.info(f'time: {t_format}')


@register_callback
class TrainFlagCallback(CallbackMixin):

    @register_to_config
    def __init__(self):
        super().__init__()

    def epoch_end(self, trainer: TrainerMixin):
        logger.info(f'epoch: {trainer.flag.epoch} step: {trainer.flag.step}')



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import copy
import functools
import inspect
import importlib
import logging
from collections import ChainMap
from dataclasses import dataclass
from tqdm import tqdm
from typing import Any, Dict, List, Optional, Union

import torch
from diffusers.utils import is_accelerate_available

if is_accelerate_available():
    import accelerate
    from accelerate import Accelerator
else:
    accelerate = None

from .config_utils import ConfigMixin
from .autosave_utils import AutoSaver, split_init_other_parameters
from .dataclass_utils import config_dataclass_wrapper
from .version import VersionInfo

# from .accelerater_utils import (
#     only_local_main_process,
#     only_main_process,
#     AccelerateMixin,
# )
# from .order_utils import ExecuteOrderMixin

logger = logging.getLogger(__name__)


def register_to_pipeline_init(init):
    @functools.wraps(init)
    def inner_init(self, *args, **kwargs):

        # Ignore private kwargs in the init.
        init_kwargs = {k: v for k, v in kwargs.items() if not k.startswith("_")}
        config_init_kwargs = {k: v for k, v in kwargs.items() if k.startswith("_")}
        if not isinstance(self, PipelineMixin):
            raise RuntimeError(
                f"`@register_to_pipeline_init` was applied to {self.__class__.__name__} init method, but this class does "
                "not inherit from `PipelineMixin`."
            )

        # Get positional arguments aligned with kwargs
        new_kwargs = {}
        signature = inspect.signature(init)
        parameters = {
            name: p.default
            for i, (name, p) in enumerate(signature.parameters.items())
            if i > 0
        }
        for arg, name in zip(args, parameters.keys()):
            new_kwargs[name] = arg

        # Then add all kwargs
        new_kwargs.update(
            {
                k: init_kwargs.get(k, default)
                for k, default in parameters.items()
                if k not in new_kwargs
            }
        )

        new_kwargs = {**config_init_kwargs, **new_kwargs}
        init(self, *args, **init_kwargs)

        for name, value in new_kwargs.items():
            self.register_save_values(**{name: value})

    return inner_init


@config_dataclass_wrapper(config_name='outputfiles.json')
@dataclass
class OutputFilesManager:
    logging_file: str | os.PathLike = "debug.log"
    logging_dir: str | os.PathLike = "logs"
    tensorboard_dir: str | os.PathLike = "tensorboard"
    wandb_dir: str | os.PathLike = "wandb"


class PipelineMixin:

    # config_name = 'pipeline_config.json'
    # record_folder: str = 'record'
    ignore_for_pipeline = set()
    output_files_manager: OutputFilesManager = OutputFilesManager()

    def __init__(self) -> None:
        super().__init__()
        self._pipeline_record = AutoSaver()

        self.register_version()
        # self.register_accelerator(accelerator_or_config)

    def register_version(self):
        self.register_save_values(_version=VersionInfo.create(clear_package=True))

    def save_pretrained(
        self,
        save_directory: str | os.PathLike,
        # safe_serialization: bool = True,
        # variant: str | None = None,
        # push_to_hub: bool = False,
        # **kwargs,
    ):
        record_path = save_directory
        self.register_version()
        self._pipeline_record.save_pretrained(record_path)

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str, **kwargs):
        # record_path = os.path.join(pretrained_model_name_or_path, cls.record_folder)
        record_path = pretrained_model_name_or_path
        # try:
        records = AutoSaver.load(record_path)
        # except Exception as e:
        #     records = {}

        kwargs = {**records, **kwargs}

        # init_parameters = inspect.signature(cls.__init__).parameters.keys()
        # init_kwargs = {k: v for k, v in kwargs.items() if k in init_parameters}
        # other_kwargs = {k: v for k, v in kwargs.items() if k not in init_parameters}
        init_kwargs, other_kwargs = split_init_other_parameters(cls, kwargs)
        # self = super().from_pretrained(pretrained_model_name_or_path, **init_kwargs)
        # print(kwargs.keys())
        # print(init_kwargs.keys())
        # print(other_kwargs.keys())
        # exit()
        self = cls(**init_kwargs)

        self.register_save_values(**other_kwargs)

        return self

    def register_save_values(self, **kwargs):

        for name, value in kwargs.items():
            if not name in self.ignore_for_pipeline:

                update_config_dict = self._pipeline_record.set(name, value)
                # self.register_to_config(**update_config_dict)

            super().__setattr__(name, value)

    def to(self, *args, **kwargs):
        dtype = kwargs.pop("dtype", None)
        device = kwargs.pop("device", None)
        dtype_arg = None
        device_arg = None
        if len(args) == 1:
            if isinstance(args[0], torch.dtype):
                dtype_arg = args[0]
            else:
                device_arg = torch.device(args[0]) if args[0] is not None else None
        elif len(args) == 2:
            if isinstance(args[0], torch.dtype):
                raise ValueError(
                    "When passing two arguments, make sure the first corresponds to `device` and the second to `dtype`."
                )
            device_arg = torch.device(args[0]) if args[0] is not None else None
            dtype_arg = args[1]
        elif len(args) > 2:
            raise ValueError(
                "Please make sure to pass at most two arguments (`device` and `dtype`) `.to(...)`"
            )

        if dtype is not None and dtype_arg is not None:
            raise ValueError(
                "You have passed `dtype` both as an argument and as a keyword argument. Please only pass one of the two."
            )

        dtype = dtype or dtype_arg

        if device is not None and device_arg is not None:
            raise ValueError(
                "You have passed `device` both as an argument and as a keyword argument. Please only pass one of the two."
            )

        device = device or device_arg

        for module in self.modules:
            is_loaded_in_8bit = (
                hasattr(module, "is_loaded_in_8bit") and module.is_loaded_in_8bit
            )

            if is_loaded_in_8bit and dtype is not None:
                logger.warning(
                    f"The module '{module.__class__.__name__}' has been loaded in 8bit and conversion to {dtype} is not yet supported. Module is still in 8bit precision."
                )

            if is_loaded_in_8bit and device is not None:
                logger.warning(
                    f"The module '{module.__class__.__name__}' has been loaded in 8bit and moving it to {dtype} via `.to()` is not yet supported. Module is still on {module.device}."
                )
            else:
                module.to(device, dtype)

        return self

    @property
    def modules(self):
        return [
            m
            for m in self._pipeline_record._auto_save_modules.values()
            if isinstance(m, torch.nn.Module)
        ]

    @property
    def device(self):

        for module in self.modules:
            if hasattr(module, "device"):
                return module.device
            for param in module.parameters():
                return param.device

        return torch.device("cpu")

    @property
    def dtype(self):

        for module in self.modules:
            if hasattr(module, "dtype"):
                return module.dtype
            for param in module.parameters():
                return param.dtype

        return torch.float32

    def tqdm(self, iterable, **kwargs):
        disable = kwargs.pop("disable", False)
        # if self.accelerator is not None:
        #     disable = not self.accelerator.is_local_main_process or disable
        return tqdm(iterable=iterable, disable=disable, **kwargs)



import os
import logging
from typing import Tuple, Dict

logger = logging.getLogger(__name__)

IMG_EXTENSIONS = (
    ".jpg",
    ".jpeg",
    ".png",
    ".ppm",
    ".bmp",
    ".pgm",
    ".tif",
    ".tiff",
    ".webp",
)


def walk_extension_files(path: str, extension: str | Tuple[str]):
    """Traverse all images in the specified path.

    Args:
        path (_type_): The specified path.

    Returns:
        List: The list that collects the paths for all the images.
    """

    file_paths = []
    for root, dirs, files in os.walk(path):
        for file in files:
            if file.endswith(extension):
                file_paths.append(os.path.join(root, file))
    logger.info(
        f'Walk path: {path}. Find {len(file_paths)} files with extension {extension}'
    )
    return file_paths


def walk_images(path: str):
    return walk_extension_files(path, suffix=IMG_EXTENSIONS)






import logging
from typing import Optional
from collections import OrderedDict

logger = logging.getLogger(__name__)

_BUILDIN_REGISTERS = {}


class Register(dict):
    def __init__(self, name: str, *args, **kwargs):
        super(Register, self).__init__(*args, **kwargs)
        self._dict = {}

        if name in _BUILDIN_REGISTERS:
            raise RuntimeError(f'The register {name} is already exist')
        logger.info(f'Create register {name}')
        _BUILDIN_REGISTERS[name] = self

        self._name = name

    @property
    def name(self):
        return self._name

    def register(self, name: Optional[str] = None):

        def _inner_register(fn):

            key = name if name is not None else fn.__name__

            if key in self._dict:
                raise RuntimeError(f'name {key} has already be register.')

            self._dict[key] = fn
            return fn

        return _inner_register

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def __str__(self):
        return str(self._dict)

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()


def list_registers():
    return list(_BUILDIN_REGISTERS.keys())


def get_register(name: str):
    if name in _BUILDIN_REGISTERS:
        return _BUILDIN_REGISTERS[name]
    return Register(name)



import os
import logging

from .constant import USER_NAME, CONFIG_HOME
from .files import load_json, safe_save_as_json

logger = logging.getLogger(__name__)

# _default_environ_init = {
#     # "WANDB_MODE": "offline",
# }

_ENVIRON_PATH = os.path.join(CONFIG_HOME, "environ.json")

_environ_config = None


def get_default_environ_init():
    global _environ_config
    if _environ_config is not None:
        return _environ_config

    if not os.path.exists(_ENVIRON_PATH):
        config = {}
        safe_save_as_json(config, _ENVIRON_PATH)
    else:
        config = load_json(_ENVIRON_PATH)

    return config


def _environ_init():

    config = get_default_environ_init()

    for key, value in config.items():
        if not key in os.environ:
            os.environ[key] = value


def set_default_environ_init(set_dict: dict):

    config = get_default_environ_init()
    for key, value in set_dict.items():
        if value is not None:
            config[key] = value
        elif key in config:
            del config[key]
    safe_save_as_json(config, _ENVIRON_PATH)
    _environ_init()



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import abc
import copy
from pathlib import Path
from typing import Literal
from ..files import safe_save_as_json, safe_save_as_yaml
from ..decorators import Register
from ..constant import ConfigSuffix

_dumper_register = Register("dumper")


class Dumper:

    def __init__(self, contents=None):
        self._contents = contents

    @property
    def contents(self):
        return self._contents

    @classmethod
    def create(cls, dtype: Literal['dict', 'list'] = 'dict'):
        cls = _dumper_register[dtype]
        if cls is None:
            raise ValueError(f'Invalid dtype: {dtype}')
        return cls()

    @abc.abstractmethod
    def _remove_dumper(self): ...

    def dump(self, dump_path: str | None = None):
        item = self._remove_dumper()
        if dump_path is None:
            if dump_path.endswith(ConfigSuffix.yaml):
                safe_save_as_yaml(item, dump_path)
            else:
                safe_save_as_json(item, dump_path)
        return item


@_dumper_register.register('dict')
class DictDumper(Dumper):

    def __init__(self, contents=None):
        super().__init__(contents=contents)
        if contents is None:
            self._contents = {}

    def update(self, key, content):
        self._contents[key] = content

    def get_subdumper(self, key, dtype: Literal['dict', 'list'] = 'dict'):
        item = self._contents.get(key)
        if item is None:
            item = self._contents[key] = self.create(dtype)
        return item

    def _remove_dumper(self):
        ret = {
            k: v._remove_dumper() if isinstance(v, Dumper) else v
            for k, v in self._contents.items()
        }
        return ret

    # def pop(self, key):
    #     return self._contents.pop(key)


@_dumper_register.register('list')
class ListDumper(Dumper):

    def __init__(self, contents=None):
        super().__init__(contents=contents)
        if contents is None:
            self._contents = []

    def update(self, content):
        self._contents.append(content)

    # def pop(self):
    #     return self._contents.pop()

    def get_subdumper(self, dtype: Literal['dict', 'list'] = 'dict'):
        item = self.create(dtype)
        self._contents.append(item)
        return item

    def _remove_dumper(self):
        ret = [
            v._remove_dumper() if isinstance(v, Dumper) else v for v in self._contents
        ]
        return ret



# Copyright 2024 Sijin Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from tqdm.auto import tqdm

from ..web.get_file import download_file
from ..web.get_json import get_json
from ..web.url_split import url_spilt
from ...check.check_args import only_one_given

logger = logging.getLogger(__name__)


def download_one_dir_from_huggingface(
    repo_id: str,
    repo_type: str,
    path: str,
    local_path: os.PathLike,
    mkdirs: bool = True,
    root_url: str = "https://huggingface.co",
    retry: int = 10,
    wait: float = 1.0,
    use_progress_bar: bool = True,
    overwrite: bool = False,
    chunk_size: int = 8192,
):
    dirname = os.path.dirname(local_path)
    if mkdirs and dirname != "":
        os.makedirs(dirname, exist_ok=True)

    _all = get_json(
        f"{root_url}/api/{repo_type}/{repo_id}/tree/main/{path}", retry=retry, wait=wait
    )

    if use_progress_bar:
        progress_bar = tqdm(
            range(len(_all)), desc=f"Downloading: {repo_id}/tree/main/{path}"
        )

    for item in _all:
        if item.get("type") == "directory":
            download_one_dir_from_huggingface(
                repo_id,
                repo_type,
                item.get("path"),
                os.path.join(local_path, item.get('path')),
                mkdirs,
                root_url,
                retry,
                wait,
                use_progress_bar,
                overwrite,
            )
        elif item.get("type", None) == "file":
            file_name = os.path.join(local_path, item.get('path'))
            if (not overwrite) and (os.path.exists(file_name)):
                logger.info(f"File {file_name} already exists, skipping download.")
                continue
            download_file(
                f"{root_url}/{repo_type}/{repo_id}/resolve/main/{item.get('path')}?download=true",
                file_name,
                mkdirs,
                chunk_size=chunk_size,
                retry=retry,
                wait=wait,
            )
        progress_bar.update(1)


def download_from_huggingface(
    url: str = None,
    repo_id: str = None,
    repo_type: str = None,
    use_mirror: bool = True,
    local_path: os.PathLike = ".download",
    retry: int = 20,
    wait: float = 5.0,
    chunk_size: int = 8192,
    continue_download: bool = True,
):
    """Downloads a dataset or model from Hugging Face, with options to use either a direct URL or a repository ID.
    The function supports downloading from a mirror site to potentially increase download speeds.

    Args:
        url (str, optional): Direct URL to the dataset or model. If specified, overrides `repo_id`. Defaults to None.
        repo_id (str, optional): Repository identifier in the format 'author/reponame'. Used if `url` is not specified.
            Defaults to None.
        repo_type (str, optional): Type of the repository ('datasets' or 'models'). Required if `repo_id` is used.
            Defaults to None.
        use_mirror (bool, optional): Whether to download from a mirror site. Defaults to True.
        local_path (os.PathLike, optional): Local filesystem path where the repository should be saved.
            Defaults to ".download".
        retry (int, optional): Number of retry attempts in case of failures. Defaults to 20.
        wait (float, optional): Wait time in seconds between retries. Defaults to 5.0.
        chunk_size (int, optional): Size of chunks to download at a time in bytes. Defaults to 8192.
        continue_download (bool, optional): If True, will continue an existing download.
            If False, will overwrite existing files. Defaults to True.

    Raises:
        ValueError: If both `url` and `repo_id` are provided, or other required parameters are missing.
    """
    if not only_one_given(url, repo_id):
        logger.error(f"Both `url` and `repo_id` are passed in.")
        raise ValueError(f"Both `url` and `repo_id` are passed in.")

    if url is not None:
        root_url, repo_type, author, repo_name, *_ = url_spilt(url)
        repo_id = f"{author}/{repo_name}"

    if use_mirror:
        root_url = "https://hf-mirror.com"
    elif url is None:
        root_url = "https://huggingface.co"

    download_one_dir_from_huggingface(
        repo_id,
        repo_type,
        "",
        local_path,
        True,
        root_url,
        retry=retry,
        wait=wait,
        chunk_size=chunk_size,
        overwrite=not continue_download,
    )



from .classifier_utils import (
    BaseImageModel,
    BaseImageClassifierOutput,
    BaseImageClassifier,
    reset_fc_layer,
    ExternalClassifier,
    BaseClassifierWrapperModel,
)
from .external import TorchvisionClassifier, ResNeStClassifier



from .base import CallbackMixin, ComposeCallback
from .avgcounter import AvgDataCallback, AvgLossCallback
from .misc import TimeCallback, TrainFlagCallback
from .checkpoint import TrainStatusCheckpointCallback, ModelCheckpointCallback



import os
import time
import argparse
import logging
import subprocess
from pynvml import *  # 或 from nvml import *
from contextlib import contextmanager

from .data import RecentKCommon

logger = logging.getLogger(__name__)


def used_mem_gb(gpu_id):
    h = nvmlDeviceGetHandleByIndex(gpu_id)
    m = nvmlDeviceGetMemoryInfo(h)
    return m.used / 1024**3


def wait_gpu_run(
    CMD: list[str] | str,
    gpu_use_num: int = 1,
    gpu_ids: str = "0-7",
    threshold_gb: float = 5,
    check_interval: float = 5,
    confirm_times: int = 3,
):
    monitor_gpus = []
    gpu_ids = gpu_ids.split(",")
    for gpu_id in gpu_ids:
        if "-" in gpu_id:
            start, end = map(int, gpu_id.split("-"))
            monitor_gpus.extend(range(start, end + 1))
        else:
            monitor_gpus.append(int(gpu_id))

    logger.info(f"Monitor GPU ID: {monitor_gpus}")
    if isinstance(CMD, str):
        CMD = [CMD]

    logger.info(f"CMD wait to run: {' '.join(CMD)}")

    try:
        # 初始化
        nvmlInit()
        _nvml_initialized = True
    except NVMLError as e:
        _nvml_initialized = False

    # with nvml_context():
    if True:

        recent_k_common = RecentKCommon(confirm_times)

        while True:
            ok_gpus = [i for i in monitor_gpus if used_mem_gb(i) < threshold_gb]

            available_gpus = recent_k_common.push(ok_gpus)

            logger.info(f"Available GPU ID: {','.join(map(str, available_gpus))}")

            if len(available_gpus) >= gpu_use_num:
                use_gpus = list(available_gpus)[:gpu_use_num]
                use_gpus_str = ",".join(map(str, use_gpus))
                logger.info(
                    f"Available GPU ID: {','.join(map(str, available_gpus))}. Run on {use_gpus_str}"
                )
                env = os.environ.copy()
                env["CUDA_VISIBLE_DEVICES"] = use_gpus_str

                if _nvml_initialized:
                    try:
                        nvmlShutdown()
                        _nvml_initialized = False  # 重置标志
                    except NVMLError as e:
                        pass

                subprocess.Popen(CMD, env=env)
                break
            else:
                print("")

            time.sleep(check_interval)


def _wait_gpu_run_handler(args):
    wait_gpu_run(
        args.cmd,
        args.gpu_use_num,
        args.gpu_ids,
        args.threshold_gb,
        args.check_interval,
        args.confirm_times,
    )


def _wait_gpu_run_register(parser: argparse._SubParsersAction):
    parser_wgr = parser.add_parser("launch", help="wait gpu run")
    parser_wgr.add_argument(
        "cmd",
        type=str,
        nargs="+",
        help="command to run",
    )
    parser_wgr.add_argument(
        "--gpu-use-num",
        '-k',
        type=int,
        default=1,
        help="number of gpu to use",
    )
    parser_wgr.add_argument(
        "--gpu-ids",
        "-g",
        type=str,
        default="0-7",
        help="gpu ids to use",
    )
    parser_wgr.add_argument(
        "--threshold-gb",
        "-s",
        type=float,
        default=5,
        help="threshold of gpu memory usage",
    )
    parser_wgr.add_argument(
        "--check-interval",
        "-i",
        type=float,
        default=5,
        help="interval of checking gpu memory usage",
    )
    parser_wgr.add_argument(
        "--confirm-times",
        "-t",
        type=int,
        default=3,
        help="number of confirm times",
    )
    parser_wgr.set_defaults(func=_wait_gpu_run_handler)



import os
import logging
from typing import Optional
from pathlib import Path

logger = logging.getLogger(__name__)


def get_safe_save_path(save_dir: str | Path, save_name: Optional[str] = None):
    save_dir = str(save_dir)
    if save_name is None:
        save_dir, save_name = os.path.split(save_dir)

    if save_dir.strip() != '':
        os.makedirs(save_dir, exist_ok=True)
        logger.info(f'Create folder {save_dir}')
    return os.path.join(save_dir, save_name)



import os
import copy
import shutil
from typing import Dict, Any, Optional

import torchvision

from . import constant
from .conversion import get_obj_from_str
from .files import safe_save_as_json
from .trainer.optimization import OptimizationConfig


class ConfigParser:
    """
    Config parser class
    """

    def __init__(
        self,
        config: Dict[str, Any],
        config_path: Optional[str] = None,
    ):
        """
        Initialize the config parser

        :param config_file: Path to the config file
        :param kwargs: Additional config parameters
        """
        self.config = config
        self.config_path = config_path
        self._visit_stack = [('/', config)]

    def save_config(self, save_path):
        """
        Save the config to a file

        :param save_path: Path to save the config file
        """

        if self.config_path is not None:
            shutil.copyfile(self.config_path, save_path)
        else:
            safe_save_as_json(self.config, save_path)

    def get(self, key: str, default: Any = None) -> Any:
        """
        Get a value from the config

        :param key: Key to get the value for
        :param default: Default value to return if the key is not found
        :return: Value from the config
        """
        try:
            return self._visit_stack[-1][1][key]
        except KeyError:
            return default

    def clone(self):
        """
        Clone the config parser

        :return: ConfigParser object
        """
        return ConfigParser(config=self.config)

    def __repr__(self):
        return f"ConfigParser(config={self.config})"

    def __str__(self) -> str:
        return self.__repr__()

    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    def cd(self, *keys: str, absolute=False):
        """
        Change directory in the config

        :param key: Key to change directory to
        :return: ConfigParser object
        """
        keys = [k for key in keys for k in key.split("/")]
        if absolute:
            self._visit_stack = [('/', self.config)]
        for key in keys:
            self._visit_stack.append((key, self._visit_stack[-1][1][key]))
        return self

    def visit(self, *keys: str, absolute=False):
        """
        Visit a key in the config

        :param key: Key to visit
        :return: ConfigParser object
        """
        keys = [k for key in keys for k in key.split("/")]
        if absolute:
            top = self.config
        else:
            top = self.top
        for key in keys:
            if key is None:
                continue
            top = top[key]
        return top

    @property
    def current_path(self):
        if len(self._visit_stack) == 1:
            return "/"
        return "/".join([key for key, _ in self._visit_stack[1:]])

    @property
    def top(self):
        """
        Get the top level config

        :return: ConfigParser object
        """
        return self._visit_stack[-1][1]

    @classmethod
    def from_yaml_config(cls, config_path: os.PathLike):
        import yaml

        with open(config_path, "r") as config_file:
            config = yaml.safe_load(config_file)
        return cls(config=config, config_path=config_path)

    @classmethod
    def from_json_config(cls, config_path: os.PathLike):
        import json

        with open(config_path, "r") as config_file:
            config = json.load(config_file)
        return cls(config=config, config_path=config_path)

    @classmethod
    def from_ini_config(cls, config_path: os.PathLike):
        import configparser

        config = configparser.ConfigParser()
        config.read(config_path)
        config = {section: dict(config.items(section)) for section in config.sections()}
        return cls(config=config, config_path=config_path)

    @classmethod
    def from_config_file(cls, config_path: os.PathLike):
        from pathlib import Path

        extension = Path(config_path).suffix.lower()
        if extension in constant.ConfigSuffix.json:
            return cls.from_json_config(config_path)
        if extension in constant.ConfigSuffix.yaml:
            return cls.from_yaml_config(config_path)
        if extension in constant.ConfigSuffix.ini:
            return cls.from_ini_config(config_path)
        raise NotImplementedError(
            f"Unknown suffix '{extension}' in path '{config_path}'."
        )

    def create_accelerator(self, subfolder="accelerator", absolute=False):
        """
        Create an accelerator config

        :param subfolder: Subfolder to create accelerator config in
        :return: accelerate.Accelerator object
        """
        import accelerate

        if subfolder not in self.config:
            return None

        config = copy.copy(self.visit(subfolder, absolute=absolute))
        if config is None:
            config = {}

        if 'dataloader_config' in config:
            dataloader_config = config['dataloader_config']
            dataloader_config = accelerate.DataLoaderConfiguration(**dataloader_config)
            config['dataloader_config'] = dataloader_config
        accelerator = accelerate.Accelerator(**config)
        return accelerator

    def create_from_name(self, name, kwargs: Dict):
        if kwargs is None:
            kwargs = {}

        builder = get_obj_from_str(name, raise_exception=True)
        return builder(**kwargs)

    def _create_transform_impl(self, name, kwargs: Dict):
        # return self.create_from_name(f'torchvision.transforms.{name}', kwargs)
        if name == "Compose":
            components = []
            for component, component_kwargs in kwargs.items():
                components.append(
                    self._create_transform_impl(component, component_kwargs)
                )
            return torchvision.transforms.Compose(components)
        else:
            return self.create_from_name(f'torchvision.transforms.{name}', kwargs)

    def create_transform(self, subfolder="transform", absolute=False):
        """
        Create a transform config

        :param subfolder: Subfolder to create transform config in
        :return: torchvision.transforms.Compose object
        """
        config = self.visit(subfolder, absolute=absolute)
        if config is None:
            return None
        return self._create_transform_impl("Compose", config)

    def create_model(self, subfolder="model", absolute=False):
        """
        Create a model config

        :param subfolder: Subfolder to create model config in
        :return: torch.nn.Module object
        """
        config = self.visit(subfolder, absolute=absolute)

        from .models import auto_model_from_pretrained, get_model_builder

        if "pretrained" in config:
            return auto_model_from_pretrained(config["pretrained"])

        builder = get_model_builder(config["name"])

        return builder(**config["kwargs"])

    def create_optimization_config(self, subfolder="optimization", absolute=False):
        """
        Create an optimization config

        :param subfolder: Subfolder to create optimization config in
        :return: torch.optim.Optimizer object
        """
        config = self.visit(subfolder, absolute=absolute)
        return OptimizationConfig(
            optimizer_name=config["optimizer"],
            optimizer_params=config["optimizer_params"],
            lr_scheduler_name=config["lr_scheduler"],
            lr_scheduler_params=config["lr_scheduler_params"],
        )



import logging
logger = logging.getLogger(__name__)

def seed_every_thing(seed: int = 0):
    """
    Set the random seed for various libraries to ensure reproducibility.

    Args:
    seed (int, Optional): The seed value to use for all random number generators. Default is 0.
    """
    _seed_python(seed)
    _seed_accelerate(seed)
    _seed_numpy(seed)
    _seed_torch(seed)

def _seed_python(seed):
    import random
    random.seed(seed)
    logger.info(f"The seed of `random` is set to {seed}.")

def _seed_torch(seed):
    try:
        import torch
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
        logger.info(f"The seed of `torch` is set to {seed}.")
    except ImportError:
        logger.warning("`torch` not found in the environment.")
        
def _seed_numpy(seed):
    try:
        import numpy
        numpy.random.seed(seed)
        logger.info(f"The seed of `numpy` is set to {seed}.")
    except ImportError:
        logger.warning("`numpy` not found in the enviroment.")

def _seed_accelerate(seed):
    try:
        import accelerate
        accelerate.utils.set_seed(seed)
        logger.info(f"The seed of `accelerate` is set to {seed}.")
    except ImportError:
        logger.warning("`accelerate` not found in the enviroment.")
            



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Union, Tuple, Literal
from dataclasses import dataclass

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

import matplotlib.pyplot as plt

from matplotlib.axes import Axes
from matplotlib import font_manager

_default_colors = [
    "#1F77B4",
    "#2CA02C",
    "#7626B4",
    "#4CCED6",
    "#FF7F0E",
    "#D62728",
    "#E317E5",
]

_default_markers = ["o", "D", "*", "p", "s", "v", "^", "X"]


from typing import Union, Tuple


def adjust_brightness(
    color: Union[str, Tuple[int, int, int]],
    factor: float,
    output_type: Literal["auto", "hex", "rgb"] = "auto",
) -> Union[str, Tuple[int, int, int]]:
    """
    Adjust the brightness of a color.
    - factor < 1: darken
    - factor > 1: lighten
    - factor = 1: no change

    Args:
        color (str | tuple): Input color, either a HEX string (e.g., "#336699")
                             or an RGB tuple (e.g., (51, 102, 153)).
        factor (float): Brightness adjustment factor.
        output_type (str): Output format, "auto", "hex", or "rgb".
                      - "auto": return in the same format as the input
                      - "hex": return HEX string
                      - "rgb": return RGB tuple
                      Defaults to "auto".

    Returns:
        str | tuple: Adjusted color in the chosen output_type.
    """
    # Convert input to RGB
    if isinstance(color, str):
        hex_color = color.lstrip('#')
        r = int(hex_color[0:2], 16)
        g = int(hex_color[2:4], 16)
        b = int(hex_color[4:6], 16)
        input_format = "hex"
    elif isinstance(color, tuple) and len(color) == 3:
        r, g, b = color
        input_format = "rgb"
    else:
        raise ValueError("color must be a HEX string or an RGB tuple")

    # Apply brightness factor
    r = min(max(0, int(r * factor)), 255)
    g = min(max(0, int(g * factor)), 255)
    b = min(max(0, int(b * factor)), 255)

    # Determine output format
    if output_type == "auto":
        output_type = input_format

    # Return in requested format
    if output_type == "hex":
        return "#{:02x}{:02x}{:02x}".format(r, g, b)
    elif output_type == "rgb":
        return (r, g, b)
    else:
        raise ValueError("output must be 'hex', 'rgb', or 'auto'")


@dataclass
class FontManager:

    font_path: str
    math_fonfamily: str = "cm"
    default_font_size: int = 14
    title_font_size: int | None = None
    axis_font_size: int | None = None
    legend_font_size: int | None = None
    tick_font_size: int | None = None

    def _get_font(self, size):
        return font_manager.FontProperties(
            fname=self.font_path,
            size=size or self.default_font_size,
            math_fontfamily=self.math_fonfamily,
        )

    def __post_init__(self):
        self.default = self._get_font(self.default_font_size)
        self.title = self._get_font(self.title_font_size)
        self.axis = self._get_font(self.axis_font_size)
        self.legend = self._get_font(self.legend_font_size)
        self.tick = self._get_font(self.tick_font_size)
        plt.rcParams.update({"font.size": self.default_font_size})


class AxesManager:

    def __init__(self, ax: Axes, font_manager: FontManager):
        self.ax = ax
        self.font_manager = font_manager
        self._init()

    def _init(self):
        self.set_xlabel(self.ax.get_xlabel())
        self.set_ylabel(self.ax.get_ylabel())
        self.set_title(self.ax.get_title())
        self.set_xticks(self.ax.get_xticks())
        self.set_yticks(self.ax.get_yticks())

    def set_xlabel(self, label):
        self.ax.set_xlabel(label, fontproperties=self.font_manager.axis)

    def set_ylabel(self, label):
        self.ax.set_ylabel(label, fontproperties=self.font_manager.axis)

    def set_title(self, title):
        self.ax.set_title(title, fontproperties=self.font_manager.title)

    def set_xticks(self, ticks):
        self.ax.set_xticks(ticks)
        self.ax.set_xticklabels(ticks, fontproperties=self.font_manager.tick)

    def set_yticks(self, ticks):
        self.ax.set_yticks(ticks)
        self.ax.set_yticklabels(ticks, fontproperties=self.font_manager.tick)

    def set_legend(self, loc, framealpha=1.0, markerscale=1, handlelength=1.7):
        self.ax.legend(
            loc=loc,
            framealpha=framealpha,
            markerscale=markerscale,
            handlelength=handlelength,
            prop=self.font_manager.legend,
        )

    def grid(self, visible=True, axis: Literal["x", "y", "both"] = "y"):
        self.ax.grid(visible=visible, axis=axis)


class Painter:

    def __init__(self, fig_size: Tuple[int, int], font_manager: FontManager):
        self.fig_size = fig_size
        self.font_manager = font_manager
        self.fig, self.ax = plt.subplots()
        self.fig.set_size_inches(*fig_size)

        self.ax_manager = AxesManager(self.ax, self.font_manager)

    def save(self, path: str, dpi: int = 600, show: bool = False):
        folder = os.path.dirname(path)
        if folder:
            os.makedirs(folder, exist_ok=True)
        self.fig.savefig(path, bbox_inches="tight", dpi=dpi)
        if show:
            plt.show()



import os
import logging
import subprocess

import torch

logger = logging.getLogger(__name__)


def _pick_gpu_by_torch() -> int | None:
    """
    Pick the GPU with the most free memory using torch.cuda.mem_get_info().
    Returns the GPU index; returns None on failure.
    """
    if not torch.cuda.is_available():
        return None
    try:
        best_i, best_free = 0, -1
        for i in range(torch.cuda.device_count()):
            with torch.cuda.device(i):
                free, total = torch.cuda.mem_get_info()  # bytes
            if free > best_free:
                best_free, best_i = free, i
        return best_i
    except Exception:
        return None


def _pick_gpu_by_nvidia_smi() -> int | None:
    """
    Fallback: call nvidia-smi to read each GPU's free memory (MiB) and pick the largest.
    """
    try:
        out = subprocess.check_output(
            ["nvidia-smi", "--query-gpu=memory.free", "--format=csv,noheader,nounits"],
            stderr=subprocess.STDOUT,
            text=True,
        )
        frees = [int(x.strip()) for x in out.strip().splitlines() if x.strip()]
        if not frees:
            return None
        # all_indices = list(range(len(frees)))
        if os.environ["CUDA_VISIBLE_DEVICES"]:
            frees = [
                frees[int(x)] for x in os.environ["CUDA_VISIBLE_DEVICES"].split(",")
            ]
        return max(list(range(len(frees))), key=lambda i: frees[i])
    except Exception:
        return None


def pick_best_gpu_index() -> int | None:
    """
    Combined strategy: prefer torch-based picking; fall back to nvidia-smi.
    """
    idx = _pick_gpu_by_torch()
    if idx is not None:
        return idx
    return _pick_gpu_by_nvidia_smi()


def auto_select_one_device():
    if torch.cuda.is_available():
        best_idx = pick_best_gpu_index()
        if best_idx is not None:
            try:
                torch.cuda.set_device(best_idx)
            except Exception:
                pass  # if set_device fails, fall back to the default CUDA device
            device = torch.device(f"cuda:{best_idx}")
            try:
                with torch.cuda.device(device):
                    free, total = torch.cuda.mem_get_info()
                logger.info(
                    f"[GPU] Auto-picked cuda:{best_idx} "
                    f"(free {free/1024/1024/1024:.1f} GB / total {total/1024/1024/1024:.1f} GB)"
                )
            except Exception:
                logger.info(f"[GPU] Auto-picked cuda:{best_idx}")
        else:
            device = torch.device("cuda")
            logger.info("[GPU] Auto-pick failed, fallback to default CUDA device.")
    else:
        device = torch.device("cpu")
        logger.info("[GPU] CUDA not available, using CPU.")



# This file includes code from the diffusers project.
# Copyright 2024 The HuggingFace Inc. team.
# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Modifications made by Chrisqcwx:
# - The ModelMixin class inherits ConfigMixin.
# - WEIGHTS_NAME and SAFETENSORS_WEIGHTS_NAME are modified to allow customization.


import os
import re
from typing import Callable, Optional, Union

import safetensors
import torch
from torchvision.transforms.functional import resize
from huggingface_hub import create_repo
from huggingface_hub.utils import validate_hf_hub_args
from torch import Tensor, nn

from diffusers import __version__
from diffusers.utils import (
    _add_variant,
    _get_model_file,
    is_accelerate_available,
    is_torch_version,
    logging,
)
from diffusers.utils.hub_utils import (
    load_or_create_model_card,
    populate_model_card,
)

from diffusers.models.modeling_utils import (
    ModelMixin as HF_ModelMixin,
    _LOW_CPU_MEM_USAGE_DEFAULT,
    is_accelerate_available,
    load_state_dict,
    load_model_dict_into_meta,
)

if is_accelerate_available():
    import accelerate
    from accelerate.utils import set_module_tensor_to_device
    from accelerate.utils.versions import is_torch_version

from ..decorators import Register

from ..auto import ConfigMixin, auto_cls_from_pretrained

logger = logging.get_logger(__name__)

_model_register = Register('model')

# register_model = _model_register.register()


def register_model(cls):
    _model_register.register()(cls)
    return cls


def list_models():
    return list(_model_register.keys())


def get_model_builder(name: str):
    return _model_register[name]


# adapt from https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/modeling_utils.py
class ModelMixin(HF_ModelMixin, ConfigMixin):

    pass


#     WEIGHTS_NAME = "pytorch_model.bin"
#     SAFETENSORS_WEIGHTS_NAME = "pytorch_model.safetensors"

#     def __init__(self):
#         super().__init__()

#         if self.WEIGHTS_NAME is None:
#             raise NotImplementedError(f'WEIGHTS_NAME is not define')
#         if self.SAFETENSORS_WEIGHTS_NAME is None:
#             raise NotImplementedError(f'SAFETENSORS_WEIGHTS_NAME is not define')

#     def save_pretrained(
#         self,
#         save_directory: Union[str, os.PathLike],
#         is_main_process: bool = True,
#         save_function: Optional[Callable] = None,
#         safe_serialization: bool = True,
#         variant: Optional[str] = None,
#         push_to_hub: bool = False,
#         **kwargs,
#     ):
#         """
#         Save a model and its configuration file to a directory so that it can be reloaded using the
#         [`~models.ModelMixin.from_pretrained`] class method.

#         Arguments:
#             save_directory (`str` or `os.PathLike`):
#                 Directory to save a model and its configuration file to. Will be created if it doesn't exist.
#             is_main_process (`bool`, *optional*, defaults to `True`):
#                 Whether the process calling this is the main process or not. Useful during distributed training and you
#                 need to call this function on all processes. In this case, set `is_main_process=True` only on the main
#                 process to avoid race conditions.
#             save_function (`Callable`):
#                 The function to use to save the state dictionary. Useful during distributed training when you need to
#                 replace `torch.save` with another method. Can be configured with the environment variable
#                 `DIFFUSERS_SAVE_MODE`.
#             safe_serialization (`bool`, *optional*, defaults to `True`):
#                 Whether to save the model using `safetensors` or the traditional PyTorch way with `pickle`.
#             variant (`str`, *optional*):
#                 If specified, weights are saved in the format `pytorch_model.<variant>.bin`.
#             push_to_hub (`bool`, *optional*, defaults to `False`):
#                 Whether or not to push your model to the Hugging Face Hub after saving it. You can specify the
#                 repository you want to push to with `repo_id` (will default to the name of `save_directory` in your
#                 namespace).
#             kwargs (`Dict[str, Any]`, *optional*):
#                 Additional keyword arguments passed along to the [`~utils.PushToHubMixin.push_to_hub`] method.
#         """
#         if os.path.isfile(save_directory):
#             logger.error(
#                 f"Provided path ({save_directory}) should be a directory, not a file"
#             )
#             return

#         os.makedirs(save_directory, exist_ok=True)

#         if push_to_hub:
#             commit_message = kwargs.pop("commit_message", None)
#             private = kwargs.pop("private", False)
#             create_pr = kwargs.pop("create_pr", False)
#             token = kwargs.pop("token", None)
#             repo_id = kwargs.pop("repo_id", save_directory.split(os.path.sep)[-1])
#             repo_id = create_repo(
#                 repo_id, exist_ok=True, private=private, token=token
#             ).repo_id

#         # Only save the model itself if we are using distributed training
#         model_to_save = self

#         # Attach architecture to the config
#         # Save the config
#         if is_main_process:
#             model_to_save.save_config(save_directory)

#         # Save the model
#         state_dict = model_to_save.state_dict()

#         weights_name = (
#             self.SAFETENSORS_WEIGHTS_NAME if safe_serialization else self.WEIGHTS_NAME
#         )
#         weights_name = _add_variant(weights_name, variant)

#         # Save the model
#         if safe_serialization:
#             safetensors.torch.save_file(
#                 state_dict,
#                 os.path.join(save_directory, weights_name),
#                 metadata={"format": "pt"},
#             )
#         else:
#             torch.save(state_dict, os.path.join(save_directory, weights_name))

#         logger.info(
#             f"Model weights saved in {os.path.join(save_directory, weights_name)}"
#         )

#         if push_to_hub:
#             # Create a new empty model card and eventually tag it
#             model_card = load_or_create_model_card(repo_id, token=token)
#             model_card = populate_model_card(model_card)
#             model_card.save(os.path.join(save_directory, "README.md"))

#             self._upload_folder(
#                 save_directory,
#                 repo_id,
#                 token=token,
#                 commit_message=commit_message,
#                 create_pr=create_pr,
#             )

#     @classmethod
#     @validate_hf_hub_args
#     def from_pretrained(
#         cls, pretrained_model_name_or_path: Optional[Union[str, os.PathLike]], **kwargs
#     ):
#         r"""
#         Instantiate a pretrained PyTorch model from a pretrained model configuration.

#         The model is set in evaluation mode - `model.eval()` - by default, and dropout modules are deactivated. To
#         train the model, set it back in training mode with `model.train()`.

#         Parameters:
#             pretrained_model_name_or_path (`str` or `os.PathLike`, *optional*):
#                 Can be either:

#                     - A string, the *model id* (for example `google/ddpm-celebahq-256`) of a pretrained model hosted on
#                       the Hub.
#                     - A path to a *directory* (for example `./my_model_directory`) containing the model weights saved
#                       with [`~ModelMixin.save_pretrained`].

#             cache_dir (`Union[str, os.PathLike]`, *optional*):
#                 Path to a directory where a downloaded pretrained model configuration is cached if the standard cache
#                 is not used.
#             torch_dtype (`str` or `torch.dtype`, *optional*):
#                 Override the default `torch.dtype` and load the model with another dtype. If `"auto"` is passed, the
#                 dtype is automatically derived from the model's weights.
#             force_download (`bool`, *optional*, defaults to `False`):
#                 Whether or not to force the (re-)download of the model weights and configuration files, overriding the
#                 cached versions if they exist.
#             resume_download (`bool`, *optional*, defaults to `False`):
#                 Whether or not to resume downloading the model weights and configuration files. If set to `False`, any
#                 incompletely downloaded files are deleted.
#             proxies (`Dict[str, str]`, *optional*):
#                 A dictionary of proxy servers to use by protocol or endpoint, for example, `{'http': 'foo.bar:3128',
#                 'http://hostname': 'foo.bar:4012'}`. The proxies are used on each request.
#             output_loading_info (`bool`, *optional*, defaults to `False`):
#                 Whether or not to also return a dictionary containing missing keys, unexpected keys and error messages.
#             local_files_only(`bool`, *optional*, defaults to `False`):
#                 Whether to only load local model weights and configuration files or not. If set to `True`, the model
#                 won't be downloaded from the Hub.
#             token (`str` or *bool*, *optional*):
#                 The token to use as HTTP bearer authorization for remote files. If `True`, the token generated from
#                 `diffusers-cli login` (stored in `~/.huggingface`) is used.
#             revision (`str`, *optional*, defaults to `"main"`):
#                 The specific model version to use. It can be a branch name, a tag name, a commit id, or any identifier
#                 allowed by Git.
#             from_flax (`bool`, *optional*, defaults to `False`):
#                 Load the model weights from a Flax checkpoint save file.
#             subfolder (`str`, *optional*, defaults to `""`):
#                 The subfolder location of a model file within a larger model repository on the Hub or locally.
#             mirror (`str`, *optional*):
#                 Mirror source to resolve accessibility issues if you're downloading a model in China. We do not
#                 guarantee the timeliness or safety of the source, and you should refer to the mirror site for more
#                 information.
#             device_map (`str` or `Dict[str, Union[int, str, torch.device]]`, *optional*):
#                 A map that specifies where each submodule should go. It doesn't need to be defined for each
#                 parameter/buffer name; once a given module name is inside, every submodule of it will be sent to the
#                 same device.

#                 Set `device_map="auto"` to have 🤗 Accelerate automatically compute the most optimized `device_map`. For
#                 more information about each option see [designing a device
#                 map](https://hf.co/docs/accelerate/main/en/usage_guides/big_modeling#designing-a-device-map).
#             max_memory (`Dict`, *optional*):
#                 A dictionary device identifier for the maximum memory. Will default to the maximum memory available for
#                 each GPU and the available CPU RAM if unset.
#             offload_folder (`str` or `os.PathLike`, *optional*):
#                 The path to offload weights if `device_map` contains the value `"disk"`.
#             offload_state_dict (`bool`, *optional*):
#                 If `True`, temporarily offloads the CPU state dict to the hard drive to avoid running out of CPU RAM if
#                 the weight of the CPU state dict + the biggest shard of the checkpoint does not fit. Defaults to `True`
#                 when there is some disk offload.
#             low_cpu_mem_usage (`bool`, *optional*, defaults to `True` if torch version >= 1.9.0 else `False`):
#                 Speed up model loading only loading the pretrained weights and not initializing the weights. This also
#                 tries to not use more than 1x model size in CPU memory (including peak memory) while loading the model.
#                 Only supported for PyTorch >= 1.9.0. If you are using an older version of PyTorch, setting this
#                 argument to `True` will raise an error.
#             variant (`str`, *optional*):
#                 Load weights from a specified `variant` filename such as `"fp16"` or `"ema"`. This is ignored when
#                 loading `from_flax`.
#             use_safetensors (`bool`, *optional*, defaults to `None`):
#                 If set to `None`, the `safetensors` weights are downloaded if they're available **and** if the
#                 `safetensors` library is installed. If set to `True`, the model is forcibly loaded from `safetensors`
#                 weights. If set to `False`, `safetensors` weights are not loaded.

#         <Tip>

#         To use private or [gated models](https://huggingface.co/docs/hub/models-gated#gated-models), log-in with
#         `huggingface-cli login`. You can also activate the special
#         ["offline-mode"](https://huggingface.co/diffusers/installation.html#offline-mode) to use this method in a
#         firewalled environment.

#         </Tip>

#         Example:

#         ```py
#         from diffusers import UNet2DConditionModel

#         unet = UNet2DConditionModel.from_pretrained("runwayml/stable-diffusion-v1-5", subfolder="unet")
#         ```

#         If you get the error message below, you need to finetune the weights for your downstream task:

#         ```bash
#         Some weights of UNet2DConditionModel were not initialized from the model checkpoint at runwayml/stable-diffusion-v1-5 and are newly initialized because the shapes did not match:
#         - conv_in.weight: found shape torch.Size([320, 4, 3, 3]) in the checkpoint and torch.Size([320, 9, 3, 3]) in the model instantiated
#         You should probably TRAIN this model on a down-stream task to be able to use it for predictions and inference.
#         ```
#         """
#         cache_dir = kwargs.pop("cache_dir", None)
#         ignore_mismatched_sizes = kwargs.pop("ignore_mismatched_sizes", False)
#         force_download = kwargs.pop("force_download", False)
#         from_flax = kwargs.pop("from_flax", False)
#         resume_download = kwargs.pop("resume_download", False)
#         proxies = kwargs.pop("proxies", None)
#         output_loading_info = kwargs.pop("output_loading_info", False)
#         local_files_only = kwargs.pop("local_files_only", None)
#         token = kwargs.pop("token", None)
#         revision = kwargs.pop("revision", None)
#         torch_dtype = kwargs.pop("torch_dtype", None)
#         subfolder = kwargs.pop("subfolder", None)
#         device_map = kwargs.pop("device_map", None)
#         max_memory = kwargs.pop("max_memory", None)
#         offload_folder = kwargs.pop("offload_folder", None)
#         offload_state_dict = kwargs.pop("offload_state_dict", False)
#         low_cpu_mem_usage = kwargs.pop("low_cpu_mem_usage", _LOW_CPU_MEM_USAGE_DEFAULT)
#         variant = kwargs.pop("variant", None)
#         use_safetensors = kwargs.pop("use_safetensors", None)

#         allow_pickle = False
#         if use_safetensors is None:
#             use_safetensors = True
#             allow_pickle = True

#         if low_cpu_mem_usage and not is_accelerate_available():
#             low_cpu_mem_usage = False
#             logger.warning(
#                 "Cannot initialize model with low cpu memory usage because `accelerate` was not found in the"
#                 " environment. Defaulting to `low_cpu_mem_usage=False`. It is strongly recommended to install"
#                 " `accelerate` for faster and less memory-intense model loading. You can do so with: \n```\npip"
#                 " install accelerate\n```\n."
#             )

#         if device_map is not None and not is_accelerate_available():
#             raise NotImplementedError(
#                 "Loading and dispatching requires `accelerate`. Please make sure to install accelerate or set"
#                 " `device_map=None`. You can install accelerate with `pip install accelerate`."
#             )

#         # Check if we can handle device_map and dispatching the weights
#         if device_map is not None and not is_torch_version(">=", "1.9.0"):
#             raise NotImplementedError(
#                 "Loading and dispatching requires torch >= 1.9.0. Please either update your PyTorch version or set"
#                 " `device_map=None`."
#             )

#         if low_cpu_mem_usage is True and not is_torch_version(">=", "1.9.0"):
#             raise NotImplementedError(
#                 "Low memory initialization requires torch >= 1.9.0. Please either update your PyTorch version or set"
#                 " `low_cpu_mem_usage=False`."
#             )

#         if low_cpu_mem_usage is False and device_map is not None:
#             raise ValueError(
#                 f"You cannot set `low_cpu_mem_usage` to `False` while using device_map={device_map} for loading and"
#                 " dispatching. Please make sure to set `low_cpu_mem_usage=True`."
#             )

#         # Load config if we don't provide a configuration
#         config_path = pretrained_model_name_or_path

#         user_agent = {
#             "diffusers": __version__,
#             "file_type": "model",
#             "framework": "pytorch",
#         }

#         # load config
#         config, unused_kwargs, commit_hash = cls.load_config(
#             config_path,
#             cache_dir=cache_dir,
#             return_unused_kwargs=True,
#             return_commit_hash=True,
#             force_download=force_download,
#             resume_download=resume_download,
#             proxies=proxies,
#             local_files_only=local_files_only,
#             token=token,
#             revision=revision,
#             subfolder=subfolder,
#             device_map=device_map,
#             max_memory=max_memory,
#             offload_folder=offload_folder,
#             offload_state_dict=offload_state_dict,
#             user_agent=user_agent,
#             **kwargs,
#         )

#         # load model
#         model_file = None

#         if True:
#             if use_safetensors:
#                 try:
#                     model_file = _get_model_file(
#                         pretrained_model_name_or_path,
#                         weights_name=_add_variant(
#                             cls.SAFETENSORS_WEIGHTS_NAME, variant
#                         ),
#                         cache_dir=cache_dir,
#                         force_download=force_download,
#                         resume_download=resume_download,
#                         proxies=proxies,
#                         local_files_only=local_files_only,
#                         token=token,
#                         revision=revision,
#                         subfolder=subfolder,
#                         user_agent=user_agent,
#                         commit_hash=commit_hash,
#                     )
#                 except IOError as e:
#                     if not allow_pickle:
#                         raise e
#                     pass
#             if model_file is None:
#                 model_file = _get_model_file(
#                     pretrained_model_name_or_path,
#                     weights_name=_add_variant(cls.WEIGHTS_NAME, variant),
#                     cache_dir=cache_dir,
#                     force_download=force_download,
#                     resume_download=resume_download,
#                     proxies=proxies,
#                     local_files_only=local_files_only,
#                     token=token,
#                     revision=revision,
#                     subfolder=subfolder,
#                     user_agent=user_agent,
#                     commit_hash=commit_hash,
#                 )

#             if low_cpu_mem_usage:
#                 # Instantiate model with empty weights
#                 with accelerate.init_empty_weights():
#                     model = cls.from_config(config, **unused_kwargs)

#                 # if device_map is None, load the state dict and move the params from meta device to the cpu
#                 if device_map is None:
#                     param_device = "cpu"
#                     state_dict = load_state_dict(model_file, variant=variant)
#                     model._convert_deprecated_attention_blocks(state_dict)
#                     # move the params from meta device to cpu
#                     missing_keys = set(model.state_dict().keys()) - set(
#                         state_dict.keys()
#                     )
#                     if len(missing_keys) > 0:
#                         raise ValueError(
#                             f"Cannot load {cls} from {pretrained_model_name_or_path} because the following keys are"
#                             f" missing: \n {', '.join(missing_keys)}. \n Please make sure to pass"
#                             " `low_cpu_mem_usage=False` and `device_map=None` if you want to randomly initialize"
#                             " those weights or else make sure your checkpoint file is correct."
#                         )

#                     unexpected_keys = load_model_dict_into_meta(
#                         model,
#                         state_dict,
#                         device=param_device,
#                         dtype=torch_dtype,
#                         model_name_or_path=pretrained_model_name_or_path,
#                     )

#                     if cls._keys_to_ignore_on_load_unexpected is not None:
#                         for pat in cls._keys_to_ignore_on_load_unexpected:
#                             unexpected_keys = [
#                                 k for k in unexpected_keys if re.search(pat, k) is None
#                             ]

#                     if len(unexpected_keys) > 0:
#                         logger.warning(
#                             f"Some weights of the model checkpoint were not used when initializing {cls.__name__}: \n {[', '.join(unexpected_keys)]}"
#                         )

#                 else:  # else let accelerate handle loading and dispatching.
#                     # Load weights and dispatch according to the device_map
#                     # by default the device_map is None and the weights are loaded on the CPU
#                     try:
#                         accelerate.load_checkpoint_and_dispatch(
#                             model,
#                             model_file,
#                             device_map,
#                             max_memory=max_memory,
#                             offload_folder=offload_folder,
#                             offload_state_dict=offload_state_dict,
#                             dtype=torch_dtype,
#                         )
#                     except AttributeError as e:
#                         # When using accelerate loading, we do not have the ability to load the state
#                         # dict and rename the weight names manually. Additionally, accelerate skips
#                         # torch loading conventions and directly writes into `module.{_buffers, _parameters}`
#                         # (which look like they should be private variables?), so we can't use the standard hooks
#                         # to rename parameters on load. We need to mimic the original weight names so the correct
#                         # attributes are available. After we have loaded the weights, we convert the deprecated
#                         # names to the new non-deprecated names. Then we _greatly encourage_ the user to convert
#                         # the weights so we don't have to do this again.

#                         if "'Attention' object has no attribute" in str(e):
#                             logger.warning(
#                                 f"Taking `{str(e)}` while using `accelerate.load_checkpoint_and_dispatch` to mean {pretrained_model_name_or_path}"
#                                 " was saved with deprecated attention block weight names. We will load it with the deprecated attention block"
#                                 " names and convert them on the fly to the new attention block format. Please re-save the model after this conversion,"
#                                 " so we don't have to do the on the fly renaming in the future. If the model is from a hub checkpoint,"
#                                 " please also re-upload it or open a PR on the original repository."
#                             )
#                             model._temp_convert_self_to_deprecated_attention_blocks()
#                             accelerate.load_checkpoint_and_dispatch(
#                                 model,
#                                 model_file,
#                                 device_map,
#                                 max_memory=max_memory,
#                                 offload_folder=offload_folder,
#                                 offload_state_dict=offload_state_dict,
#                                 dtype=torch_dtype,
#                             )
#                             model._undo_temp_convert_self_to_deprecated_attention_blocks()
#                         else:
#                             raise e

#                 loading_info = {
#                     "missing_keys": [],
#                     "unexpected_keys": [],
#                     "mismatched_keys": [],
#                     "error_msgs": [],
#                 }
#             else:
#                 model = cls.from_config(config, **unused_kwargs)

#                 state_dict = load_state_dict(model_file, variant=variant)
#                 model._convert_deprecated_attention_blocks(state_dict)

#                 model, missing_keys, unexpected_keys, mismatched_keys, error_msgs = (
#                     cls._load_pretrained_model(
#                         model,
#                         state_dict,
#                         model_file,
#                         pretrained_model_name_or_path,
#                         ignore_mismatched_sizes=ignore_mismatched_sizes,
#                     )
#                 )

#                 loading_info = {
#                     "missing_keys": missing_keys,
#                     "unexpected_keys": unexpected_keys,
#                     "mismatched_keys": mismatched_keys,
#                     "error_msgs": error_msgs,
#                 }

#         if torch_dtype is not None and not isinstance(torch_dtype, torch.dtype):
#             raise ValueError(
#                 f"{torch_dtype} needs to be of type `torch.dtype`, e.g. `torch.float16`, but is {type(torch_dtype)}."
#             )
#         elif torch_dtype is not None:
#             model = model.to(torch_dtype)

#         model.register_to_config(_name_or_path=pretrained_model_name_or_path)

#         # Set model in evaluation mode to deactivate DropOut modules by default
#         model.eval()
#         if output_loading_info:
#             return model, loading_info

#         return model


def auto_model_from_pretrained(path: str, **kwargs):

    return auto_cls_from_pretrained(_model_register, ModelMixin, path, **kwargs)



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
from typing import Dict

import torch
from torch import compile
from torch.nn.parallel import DataParallel, DistributedDataParallel

import diffusers
from diffusers.utils import is_accelerate_available, is_torch_version

from .autosave_utils import auto_register_save_load

logger = logging.getLogger(__name__)

if is_accelerate_available():
    import accelerate
    from accelerate import Accelerator, DataLoaderConfiguration

    auto_register_save_load(DataLoaderConfiguration)

    auto_register_save_load(
        Accelerator,
        ignore_list=[
            'dispatch_batches',
            'split_batches',
            'even_batches',
            'use_seedable_sampler',
        ],
    )
else:
    accelerate = None


def is_compiled_module(module):
    """
    Check whether the module was compiled with torch.compile()
    """
    if is_torch_version("<", "2.0.0") or not hasattr(torch, "_dynamo"):
        return False
    return isinstance(module, torch._dynamo.eval_frame.OptimizedModule)


def only_local_main_process(fn):

    @functools.wraps(fn)
    def inner_fn(self: AccelerateMixin, *args, **kwargs):
        self.wait_for_everyone()
        if self.is_local_main_process:
            return fn(self, *args, **kwargs)
        self.wait_for_everyone()

    return inner_fn


def only_main_process(fn):

    @functools.wraps(fn)
    def inner_fn(self: AccelerateMixin, *args, **kwargs):
        self.wait_for_everyone()
        if self.is_main_process:
            return fn(self, *args, **kwargs)
        self.wait_for_everyone()

    return inner_fn


class AccelerateMixin:

    _use_ddp: bool = False
    _accelerator: "Accelerator" = None

    def register_accelerator(
        self, accelerator_or_config: Dict | "Accelerator" = None
    ) -> None:
        if accelerate is None or accelerator_or_config is None:
            if accelerator_or_config is not None:
                logger.warning(
                    "accelerate is not installed, so the accelerator_config will be ignored."
                )
            self._accelerator = None
            self._use_ddp = False
        else:
            if not isinstance(accelerator_or_config, Accelerator):
                self._accelerator = Accelerator(**accelerator_or_config)
            else:
                self._accelerator = accelerator_or_config
            self._use_ddp = True

    @property
    def accelerator(self) -> "Accelerator":
        return self._accelerator

    @property
    def use_ddp(self) -> bool:
        return self._use_ddp

    @property
    def is_local_main_process(self) -> bool:
        if self.accelerator is None:
            return True
        return self.accelerator.is_local_main_process

    @property
    def is_main_process(self) -> bool:
        if self.accelerator is None:
            return True
        return self.accelerator.is_main_process

    def wait_for_everyone(self):
        if self.accelerator is not None:
            self.accelerator.wait_for_everyone()

    def accelerator_prepare(self, *args, device_placement=None):
        if self.accelerator is not None:
            self.accelerator.prepare(*args, device_placement=device_placement)

    def unwrap_model(self, model):
        if self.accelerator is not None:
            return self.accelerator.unwrap_model(model)

        if is_compiled_module(model):
            model = model._orig_mod

        if isinstance(model, (DataParallel, DistributedDataParallel)):
            model = model.module

        return model

    def accelerator_log(self, values: dict, step: int | None = None):
        if self.accelerator is not None:
            self.accelerator.log(values, step=step)



import abc
import os
from typing import Literal

import torch
from diffusers.configuration_utils import register_to_config

from .base import CallbackMixin, register_callback
from ..trainer_utils import TrainerMixin
from ..constant import TRAINER_STATUS_NAME


class CheckpointCallbackMixin(CallbackMixin):

    def __init__(self, interval: int, level: Literal['epoch', 'step'] = 'epoch'):
        super().__init__()
        self.interval = interval

        self.level = level
        if level == 'epoch':
            self.on_train_epoch_end = self._epoch_end
        elif level == 'step':
            self.on_train_step_end = self._step_end
        else:
            raise RuntimeError(f'The level must be epoch or step')

    def get_value(self, trainer: TrainerMixin):
        return trainer.flag.epoch if self.level == 'epoch' else trainer.flag.step

    @abc.abstractmethod
    def save_checkpoint(self, trainer: TrainerMixin):
        pass

    def _step_end(self, trainer: TrainerMixin):
        if trainer.is_local_main_process and trainer.flag.step % self.interval == 0:
            self.save_checkpoint(trainer)

    def _epoch_end(self, trainer: TrainerMixin):
        if trainer.is_local_main_process and trainer.flag.epoch % self.interval == 0:
            self.save_checkpoint(trainer)

    def on_train_end(self, trainer: TrainerMixin):
        if trainer.is_local_main_process:
            self.save_checkpoint(trainer)


@register_callback
class ModelCheckpointCallback(CheckpointCallbackMixin):

    @register_to_config
    def __init__(self, interval: int, level: Literal['epoch'] | Literal['step'] = 'epoch'):
        super().__init__(interval, level)

    def save_checkpoint(self, trainer: TrainerMixin):
        trainer._save_checkpoint()

    def on_train_end(self, trainer: TrainerMixin):
        self.save_checkpoint(trainer)


@register_callback
class TrainStatusCheckpointCallback(CheckpointCallbackMixin):

    @register_to_config
    def __init__(self, interval: int, level: Literal['epoch'] | Literal['step'] = 'epoch'):
        super().__init__(interval, level)

    def save_checkpoint(self, trainer: TrainerMixin):
        trainer._save_training_status()



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
from dataclasses import dataclass
from typing import Dict, Union, Any, Optional, AnyStr

import torch
from diffusers.optimization import get_scheduler as diffusers_get_scheduler

from ..conversion import get_obj_from_str
from ..decorators import Register
from ..files import safe_save_as_json, load_json, get_safe_save_path
from ..auto import config_dataclass_wrapper


logger = logging.getLogger(__name__)

optimizer_register = Register('optimizer')
lr_shceduler_register = Register('lr_scheduler')


def _search(
    register: Register,
    name: Optional[AnyStr] = None,
    first_input: Any = None,
    build_params: Optional[Dict] = None,
    source: Optional[str] = None,
):
    if name is None:
        return None

    if source is not None and source not in register.keys():
        raise RuntimeError(f'Source {source} is not valid')

    if build_params is None:
        build_params = {}

    search_keys = [source] if source is not None else register.keys()
    for src in search_keys:
        cls_builder = register[src]
        instance = cls_builder(name, first_input, build_params)
        if instance is not None:
            logger.info(f'get {register.name} from {src}')
            return instance

    raise RuntimeError(f'Name {name} can not be found for {register.name}.')


@optimizer_register.register('torch')
def _get_torch_optimizer(name: str, params, build_params):
    if not name.startswith('torch.optim.'):
        name = f'torch.optim.{name}'
    optimizer_cls = get_obj_from_str(name)
    return optimizer_cls(params, **build_params)


@lr_shceduler_register.register('torch')
def _get_torch_lr_scheduler(name: str, optimizer, kwargs):
    lr_scheduler_cls = get_obj_from_str(f'torch.optim.lr_scheduler.{name}')
    if lr_scheduler_cls is None:
        return None
    return lr_scheduler_cls(optimizer, **kwargs)


@lr_shceduler_register.register('diffusers')
def _get_diffusers_lr_scheduler(name: str, optimizer, kwargs):
    try:
        lr_scheduler = diffusers_get_scheduler(name, optimizer, **kwargs)
    except:
        return None

    return lr_scheduler


def get_optimizer(name, params, optimizer_params):

    return _search(optimizer_register, name, params, optimizer_params)


def get_lr_scheduler(name, optimizer, lr_scheduler_params):

    return _search(lr_shceduler_register, name, optimizer, lr_scheduler_params)


@config_dataclass_wrapper('optimization_config.json')
@dataclass
class OptimizationConfig:
    optimizer_name: str
    optimizer_params: Dict = None
    lr_scheduler_name: str = None
    lr_scheduler_params: Dict = None


class SingleOptimizationManager:

    save_name = 'optimization.pt'

    def __init__(
        self,
        params,
        config: OptimizationConfig,
    ):
        self.params = params
        self._config = config
        self.optimizer = get_optimizer(
            config.optimizer_name, params, config.optimizer_params
        )
        self.lr_scheduler = get_lr_scheduler(
            config.lr_scheduler_name, self.optimizer, config.lr_scheduler_params
        )

    def zero_grad(self):
        self.optimizer.zero_grad()

    def optimizer_step(self):
        self.optimizer.step()

    def lr_scheduler_step(self):
        if self.lr_scheduler is not None:
            self.lr_scheduler.step()

    def state_dict(self):
        state_dict = {
            'optimizer': self.optimizer.state_dict(),
            'lr_scheduler': (
                self.lr_scheduler.state_dict()
                if self.lr_scheduler is not None
                else None
            ),
        }

        return state_dict

    def load_state_dict(self, state_dict):
        self.optimizer.load_state_dict(state_dict['optimizer'])
        if self.lr_scheduler is not None and 'lr_scheduler' in state_dict:
            self.lr_scheduler.load_state_dict(state_dict['lr_scheduler'])

    def load_state_dict_from_file(self, save_directory: str):
        if not save_directory.endswith(self.save_name):
            save_path = os.path.join(save_directory, self.save_name)
        else:
            save_path = save_directory
        state_dict = torch.load(save_path, map_location=next(self.params).device)
        self.load_state_dict(state_dict)

    def save_state_dict_to_file(self, save_directory: str):
        if not save_directory.endswith(self.save_name):
            save_path = os.path.join(save_directory, self.save_name)
        else:
            save_path = save_directory
        save_path = get_safe_save_path(save_path)
        torch.save(self.state_dict(), save_path)

    def __iter__(self):
        # 定义迭代器，返回两个值
        yield self.optimizer
        yield self.lr_scheduler

    def accelerator_prepare(self, accelerator: "Accelerator", device_placement=True):
        self.optimizer = accelerator.prepare(
            self.optimizer, device_placement=device_placement
        )
        if self.lr_scheduler is not None:
            self.lr_scheduler = accelerator.prepare(
                self.lr_scheduler, device_placement=device_placement
            )


class OptimizationManager:

    optimization_group: Dict[str, SingleOptimizationManager] = {}

    def __init__(self) -> None:
        pass

    def register_optimization(
        self,
        tag,
        params,
        optimization_config,
    ):
        _optimize_manager = SingleOptimizationManager(params, optimization_config)
        self.optimization_group[tag] = _optimize_manager

        return _optimize_manager

    def get_optimization(self, tag):
        return self.optimization_group[tag]

    def __getitem__(self, tag):
        return self.get_optimization(tag)

    def get_optimizer(self, tag):
        return self.optimization_group[tag].optimizer

    def get_lr_scheduler(self, tag):
        return self.optimization_group[tag].lr_scheduler

    def state_dict(self, tag: Optional[str] = None):
        if tag is None:
            return {
                tag: manager.state_dict()
                for tag, manager in self.optimization_group.items()
            }
        else:
            return self.optimization_group[tag].state_dict()

    def load_state_dict_from_file(self, save_directory: str):
        for tag, manager in self.optimization_group.items():
            tag_directory = os.path.join(save_directory, tag)
            if not os.path.exists(tag_directory):
                logger.warning(
                    f'No save directory found for optimization {tag}. Automatically ignore it.'
                )
                continue
            manager.load_state_dict_from_file(tag_directory)

    def save_state_dict_to_file(self, save_directory: str):
        for tag, manager in self.optimization_group.items():
            tag_directory = os.path.join(save_directory, tag)
            manager.save_state_dict_to_file(tag_directory)

    def accelerator_prepare(self, accelerator: "Accelerator", device_placement=True):
        for manager in self.optimization_group.values():
            manager.accelerator_prepare(accelerator, device_placement=device_placement)

    def zero_grad(self, tag: Optional[str] = None):
        if tag is None:
            for manager in self.optimization_group.values():
                manager.optimizer.zero_grad()
        else:
            self.optimization_group[tag].optimizer.zero_grad()

    def optimizer_step(self, tag: Optional[str] = None):
        if tag is None:
            for manager in self.optimization_group.values():
                manager.optimizer.step()
        else:
            self.optimization_group[tag].optimizer.step()

    def lr_scheduler_step(self, tag: Optional[str] = None):
        if tag is None:
            for manager in self.optimization_group.values():
                if manager.lr_scheduler is not None:
                    manager.lr_scheduler.step()
        else:
            if self.optimization_group[tag].lr_scheduler is not None:
                self.optimization_group[tag].lr_scheduler.step()



import re


def split_with_punctuation(s: str, split_pattern):
    # 匹配以 .,!? 中任意一个符号结尾的片段
    parts = re.findall(r'.*?[{}](?=\s|$)'.format(split_pattern), s)

    remainder = s[sum(len(p) for p in parts) :]
    if remainder:
        parts.append(remainder)
    return parts



import abc
import datetime
import functools
import logging
import os
from dataclasses import dataclass
from typing import Dict

import torch
from diffusers.configuration_utils import ConfigMixin, register_to_config
from torch.utils.tensorboard import SummaryWriter

from . import constant
from ..logger.logging import init_logger
from ..random._seeds import seed_every_thing

logger = logging.getLogger(__name__)

class TrainingConfig:
    def __init__(
        self,
        max_epoches: int = None,
        max_steps: int = None,
        checkpointing_epoches: int = None,
        checkpointing_steps: int = None,
        validation_epoches: int = None,
        validation_steps: int = None,
        saving_epoches: int = None,
        saving_steps: int = None,
        watching_epoches: int = None,
        watching_steps: int = None
    ) -> None:
        self.max_epoches = max_epoches
        self.max_steps = max_steps
        if (max_epoches is None) and (max_steps is None):
            logger.warning(f"Both `max_epochs` and `max_steps` are None. "
                           f"`max_epochs` is automatically set to 10000.")
            self.max_epoches = 10000
        elif (max_epoches is not None) and (max_steps is not None):
            logger.warning(f"Both `max_epochs` and `max_first` are given. "
                           f"Training will end when either limit is reached.")
        
        self.checkpointing_epoches = checkpointing_epoches
        self.checkpointing_steps = checkpointing_steps
        if (checkpointing_epoches is None) and (checkpointing_steps is None):
            logger.warning(f"Both `checkpointing_epochs` and `checkpointing_steps` are None. "
                           f"No checkpoints will be saved during the training.")
        elif (checkpointing_epoches is not None) and (checkpointing_steps is not None):
            logger.warning(f"Both `checkpointing_epochs` and `checkpointing_steps` are given. "
                           f"All checkpoints meeting the criteria will be saved.")
        
        self.validation_epoches = validation_epoches
        self.validation_steps = validation_steps
        if (validation_epoches is None) and (validation_steps is None):
            logger.warning(f"Both `validation_epochs` and `validation_steps` are None. "
                           f"No validation will be performed during the training.")
        elif (validation_epoches is not None) and (validation_steps is not None):
            logger.warning(f"Both `validation_epochs` and `validation_steps` are given. "
                           f"All validation meeting the criteria will be performed.")
        
        self.saving_epoches = saving_epoches
        self.saving_steps = saving_steps
        if (saving_epoches is None) and (saving_steps is None):
            logger.warning(f"Both `saving_epochs` and `saving_steps` are None. "
                           f"No states will be saved during the training.")
        elif (saving_epoches is not None) and (saving_steps) is not None:
            logger.warning(f"Both `saving_epochs` and `saving_steps` are given. "
                           f"All states meeting the criteria will be saved.")
        
        self.watching_epoches = watching_epoches
        self.watching_steps = watching_steps
        if (watching_epoches is None) and (watching_steps is None):
            logger.warning(f"Both `saving_epochs` and `saving_steps` are None. "
                           f"No variables will be saved during the training.")
        elif (watching_epoches is not None) and (watching_steps is not None):
            logger.warning(f"Both `saving_epochs` and `saving_steps` are given. "
                           f"all variables meeting the criteria will be saved.")
    

@dataclass
class TrainingFlag:
    distributed_training: bool
    device_id: int
    
    @property
    def is_main_process(self) -> bool:
        if self.distributed_training:
            return self.device_id == 0
        else:
            return True
    


# TODO: 请测试
class TrainerMixin(abc.ABC, ConfigMixin):
    
    config_name = "config.json"
    logger_name = "training.log"
    tensorboard_subfolder = "tensorboard"
    checkpoints_subfolder = "checkpoints"
    last_training_status_subfolder = "last"
    
    ignore_for_config = ["_reload"]
    
    @register_to_config
    def __init__(
        self,
        output_dir_config: dict,
        training_config: dict,
        seed: int,
        distributed_training: bool,
        device_id: int,
        *,
        _reload: bool = False,
        _output_path: os.PathLike = None,
    ) -> None:
        super().__init__()
        self.training_flag = TrainingFlag(distributed_training, device_id)
        self.is_main_process = self.training_flag.is_main_process
        if self.is_main_process:
            self._set_output_dir(**output_dir_config, _reload=_reload, _output_path=_output_path)
        self._set_seed(seed)
        self._set_training_config(**training_config)
        
        self.current_epoch = 0
        self.current_step = 0
        self.training_dataset = None
        self.training_dataloader = None
        
    
    def _set_output_dir(
        self,
        root: os.PathLike = ".outputs",
        name: os.PathLike = None,
        use_timestamp: bool = True,
        logger_config: Dict[str, str] = None,
        *,
        _reload: bool = False,
        _output_path: os.PathLike = None
    ):
        if _reload:
            self.output_dir = _output_path
        else:
            if name is None:
                path = root
            else:
                path = os.path.join(root, name)
            
            if use_timestamp:
                now = datetime.datetime.now().strftime("%Y-%m-%d,%H:%M:%S,%f")[:-3]
                path = os.path.join(path, now)
            
            os.makedirs(path, exist_ok=True)
            self.output_dir = path
        
        init_logger(
            file_name=os.path.join(self.output_dir, self.logger_name),
            **logger_config
        )
        
        os.makedirs(os.path.join(self.output_dir, self.tensorboard_subfolder), exist_ok=True)
        self.writer = SummaryWriter(os.path.join(self.output_dir, self.tensorboard_subfolder))
        
        self.save_config(save_directory=self.output_dir, push_to_hub=False)
    
    
    def _set_seed(
        self,
        seed: int
    ):
        self.seed = seed
        seed_every_thing(seed)
        
    
    def _set_training_config(
        self,
        training_config: Dict
    ):
        self.training_config = TrainingConfig(**training_config)
    
    @abc.abstractmethod
    def auto_init(self, reload: bool = False):
        raise NotImplementedError("Please implement the `auto_init` method.")
    
    @abc.abstractmethod
    def auto_set_to_train_mode(self):
        raise NotImplementedError("Please implement the `auto_set_to_train_mode` method.")
    
    @abc.abstractmethod
    def auto_set_to_eval_mode(self):
        raise NotImplementedError("Please implement the `auto_set_to_eval_mode` method.")
    
    @abc.abstractmethod
    def auto_save_checkpoint(self):
        raise NotImplementedError("Please implement the `auto_save_checkpoint` method.")
    
    @abc.abstractmethod
    def auto_save_training_status(self):
        raise NotImplementedError("Please implement the `auto_save_training_status` method.")
    
    @abc.abstractmethod
    def auto_validation(self):
        raise NotImplementedError("Please implement the `auto_validation` method.")
    
    @abc.abstractmethod
    def auto_watching(self):
        raise NotImplementedError("Please implement the `auto_watching` method.")
    
    @abc.abstractmethod
    def auto_load_status(self, checkpoint_path):
        raise NotImplementedError("Please implement the `auto_load_status` method.")
    
    @classmethod
    def from_yaml_config(cls, config_path: os.PathLike, _reload: bool = False):
        import yaml
        with open(config_path, "r") as config_file:
            config = yaml.safe_load(config_file)
            if _reload:
                config["_reload"] = _reload
                config["_output_path"] = os.path.dirname(config_path)
        return cls.from_config(config)
    
    @classmethod
    def from_json_config(cls, config_path: os.PathLike, _reload: bool = False):
        import json
        with open(config_path, "r") as config_file:
            config = json.load(config_file)
            if _reload:
                config["_reload"] = _reload
                config["_output_path"] = os.path.dirname(config_path)
        return cls.from_config(config)
    
    @classmethod
    def from_ini_config(cls, config_path: os.PathLike, _reload: bool = False):
        import configparser
        config = configparser.ConfigParser()
        config.read(config_path)
        config = {section: dict(config.items(section)) for section in config.sections()}
        if _reload:
                config["_reload"] = _reload
                config["_output_path"] = os.path.dirname(config_path)
        return cls.from_config(config)
    
    @classmethod
    def from_py_config(
        cls, 
        config_path: os.PathLike = None, 
        config_module: str = None,
        _reload: bool = False
    ):
        # TODO: 完善从 py 文件读取配置文件的功能.
        # TODO: 太恶心了 我写不动 谁爱用这个功能谁写.
        raise NotImplementedError
        if (config_path is None) and (config_module is None):
            raise ValueError("Cannot init form config with both `config_path` and `config_module` are None.")
        if (config_path is not None) and (config_module is not None):
            logger.warning(f"Both `config_path` and `config_module` are given. "
                           f"`config_path` is used by default.")
        if config_path is not None:
            pass
    
    @classmethod
    def from_config_file(cls, config_path: os.PathLike, *, _reload: bool = False):
        from pathlib import Path
        extension = Path(config_path).suffix.lower()
        if extension in constant.ConfigSuffix.json:
            return cls.from_json_config(config_path, _reload=_reload)
        if extension in constant.ConfigSuffix.yaml:
            return cls.from_yaml_config(config_path, _reload=_reload)
        if extension in constant.ConfigSuffix.ini:
            return cls.from_ini_config(config_path, _reload=_reload)
        raise NotImplementedError(f"Unknown suffix '{extension}' in path '{config_path}'.")
            
    @classmethod
    def from_output_dir(cls, dir_path: os.PathLike):
        config_path = os.path.join(dir_path, cls.config_name)
        self = cls.from_config_file(config_path, _reload=True)
        self.auto_load_status()
        
        logger.info(f"Warm up for epoch={self.current_epoch} and step={self.current_step}.")
        if self.current_step == 0 and self.current_epoch == 0:
            return self
        _step = 0
        _epoch = 0
        while True:
            for _ in self.training_dataloader:
                _step += 1
                if _step == self.current_step:
                    if _epoch == self.current_epoch:
                        return self
                    else:
                        raise RuntimeError("`current_epoch` and `current_step` mismatch.")
            _epoch += 1
            

def register_to_run_one_epoch(one_epoch_func: function, only_training: bool = False):
    
    @functools.wraps(one_epoch_func)
    def run_one_epoch(self: TrainerMixin, *args, **kwargs):
        result = one_epoch_func(*args, **kwargs)
        self.current_epoch += 1
        
        if only_training:
            return result
        
        self.auto_set_to_eval_mode()
        
        with torch.no_grad():
            if self.training_config.checkpointing_epoches is not None:
                if self.current_epoch % self.training_config.checkpointing_epoches == 0:
                    self.auto_save_checkpoint()
            
            if self.training_config.saving_epoches is not None:
                if self.current_epoch % self.training_config.saving_epoches == 0:
                    self.auto_save_training_status()
            
            if self.training_config.validation_epoches is not None:
                if self.current_epoch % self.training_config.validation_epoches == 0:
                    self.auto_validation()
                
            if self.training_config.watching_epoches is not None:
                if self.current_epoch % self.training_config.watching_epoches == 0:
                    self.auto_watching()
        
        self.auto_set_to_train_mode()
        
        return result
    
    return run_one_epoch


def register_to_run_one_iteration(one_iteration_func: function, only_training: bool = False):
    
    @functools.wraps(one_iteration_func)
    def run_one_iteration(self: TrainerMixin, *args, **kwargs):
        result = one_iteration_func(*args, **kwargs)
        self.current_step += 1
        
        if only_training:
            return result
        
        self.auto_set_to_eval_mode()
        
        with torch.no_grad():
            if self.training_config.checkpointing_steps is not None:
                if self.current_step % self.training_config.checkpointing_steps == 0:
                    self.auto_save_checkpoint()
            
            if self.training_config.saving_steps is not None:
                if self.current_step % self.training_config.saving_steps == 0:
                    self.auto_save_training_status()
            
            if self.training_config.validation_steps is not None:
                if self.current_step % self.training_config.validation_steps == 0:
                    self.auto_validation()
                
            if self.training_config.watching_steps is not None:
                if self.current_step % self.training_config.watching_steps == 0:
                    self.auto_watching()
        
        self.auto_set_to_train_mode()
        
        return result
    
    return run_one_iteration


import json
import logging

logger = logging.getLogger(__name__)


def load_yaml(path: str):

    import yaml

    with open(path, 'r') as f:
        data = yaml.safe_load(f)

    return data


def load_torchobj(path: str):

    import torch

    return torch.load(path, map_location='cpu')


def load_csv(path: str):
    import pandas as pd

    return pd.read_csv(path)


def load_json(path: str):

    with open(path, 'r') as f:
        data = json.load(f)

    return data


def load_pickle(path: str):

    import pickle

    with open(path, 'rb') as f:
        data = pickle.load(f)

    return data



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import functools
from typing import Callable, Optional

import torch
from tqdm import tqdm
from diffusers.utils.outputs import BaseOutput

from ..format.format import format_as_split_line


def _is_namedtuple(obj):
    return (
        isinstance(obj, tuple) and hasattr(obj, "_asdict") and hasattr(obj, "_fields")
    )


def _gather(outputs, dim=0):
    """Gather the input data.

    Args:
        outputs (_type_): The data to gather.
        dim (int, optional): The specified dimension used when the type of input data is torch.Tensor. Defaults to 0.
    """

    def gather_map(outputs):
        out = outputs[0]
        if isinstance(out, torch.Tensor):
            return torch.cat(outputs, dim=dim)
        if out is None:
            return None
        if isinstance(out, (list, tuple)) and isinstance(out[0], str):
            return list(functools.reduce(lambda x, y: x + y, outputs))
        if isinstance(out, BaseOutput):
            return type(out)(*gather_map([d.to_tuple() for d in outputs]))
        if isinstance(out, dict):
            if not all(len(out) == len(d) for d in outputs):
                raise ValueError('All dicts must have the same number of keys')
            return type(out)((k, gather_map([d[k] for d in outputs])) for k in out)
        if _is_namedtuple(out):
            return type(out)._make(map(gather_map, zip(*outputs)))
        return type(out)(map(gather_map, zip(*outputs)))

    try:
        res = gather_map(outputs)
    finally:
        gather_map = None
    return res


def batch_apply(
    fn: Callable,
    *inputs,
    batch_size: int,
    description: Optional[str] = None,
    use_tqdm: bool = False,
    **other_input_kwargs,
):
    """Apply the given function to input data by the specified batch size.

    Args:
        fn (Callable): The given function.
        *inputs: The collected input data.
        batch_size (int): The specified batch size.
        description (Optional[str], optional): The content to print when processing the input data. Defaults to None.
        use_tqdm (bool, optional): Determine whether to use tqdm when printing. Defaults to False.
    """

    def _check_valid(inputs):
        if len(inputs) == 0:
            return
        lens = []
        for i, inp in enumerate(inputs):
            try:
                lens.append(len(inp))
            except:
                raise RuntimeError(f'the {i} inputs have no attr `len`')
        valid_len = lens[0]
        if not all(map(lambda x: x == valid_len, lens)):
            raise RuntimeError('lengths of all inputs are not the same')

    _check_valid(inputs)

    total_len = len(inputs[0])

    results = []
    starts = list(range(0, total_len, batch_size))
    iter_times = len(starts)

    if use_tqdm:
        if description is not None:
            print(format_as_split_line(description))
        starts = tqdm(starts, leave=False)

    for i, start in enumerate(starts, start=1):

        if description is not None and not use_tqdm:
            print(format_as_split_line(f'{description}: {i} / {iter_times}'))

        end = min(total_len, start + batch_size)
        res = fn(*[p[start:end] for p in inputs], **other_input_kwargs)
        # print(res.device)
        results.append(res)
    return _gather(results)



# Copyright 2024 Sijin Yu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import logging
import os
import sys

logger = logging.getLogger(__name__)


def is_huggingface_hub_imported():
    return "huggingface_hub" in sys.modules


def set_huggingface_mirror(url=None):
    if is_huggingface_hub_imported():
        logger.error("`huggingface_hub` is imported before setting the mirror.")

    url = "https://hf-mirror.com" if url is None else url
    os.environ["HF_ENDPOINT"] = url
    logger.info(f"The `huggingface_hub` mirror is set to '{url}'")


def hf_hub_download(mirror_url: str = None, **kwargs):
    set_huggingface_mirror(mirror_url)
    import huggingface_hub

    return huggingface_hub.hf_hub_download(**kwargs)


def load_dataset(mirror_url: str = None, **kwargs):
    set_huggingface_mirror(mirror_url)
    import datasets

    return datasets.load_dataset(**kwargs)



# Copyright 2024 Sijin Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Dict

import torch.distributed
from torch.utils.data import Dataset, DataLoader, RandomSampler, SequentialSampler
from torch.utils.data.distributed import DistributedSampler


@dataclass
class DataLoaderConfig:
    num_workers: int = 1
    pin_memory: bool = True
    drop_last: bool = False
    prefetch_factor: int = 2
    presistent_workers: bool = True


def getDataLoader(
    dataset: Dataset,
    batch_size: int,
    config: Dict,
    shuffle: bool = True,
    device: str = "",
):

    sampler = None

    dataloader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        sampler=sampler,
        num_workers=config.get("num_workers", 0),
        pin_memory=config.get("pin_memory", False),
        drop_last=config.get("drop_last", False),
        prefetch_factor=config.get("prefetch_factor", None),
        persistent_workers=config.get("presistent_workers", False),
        pin_memory_device=device,
    )

    return dataloader



# Copyright 2025 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import abc
import copy
import json
import inspect
import logging
import functools
from typing import Callable
from collections import ChainMap
from typing import List

import torch

from ..files import safe_save_as_json, load_json
from ..conversion import get_obj_from_str

logger = logging.getLogger(__name__)


def get_init_parameters(cls, pop_kwargs=False):
    result = dict(inspect.signature(cls.__init__).parameters)
    result.pop("self", None)
    if pop_kwargs:
        result.pop("kwargs", None)
    return result


def split_init_other_parameters(cls, parameters):
    init_parameters = get_init_parameters(cls, pop_kwargs=True).keys()
    init_results = {k: v for k, v in parameters.items() if k in init_parameters}
    other_results = {k: v for k, v in parameters.items() if k not in init_parameters}
    return init_results, other_results


def auto_create_cls(cls, config, **kwargs):
    kwargs = {**{k: v for k, v in config.items() if k not in kwargs}, **kwargs}
    init_kwargs, _ = split_init_other_parameters(cls, kwargs)
    return cls(**init_kwargs)


def _register_fn(_method_name, fn: str | Callable, cls=None):

    def _inner_wrapper(cls):
        nonlocal fn
        if isinstance(fn, str):
            fn = getattr(cls, fn, None)
            if fn is None:
                raise ValueError(f"Cannot find method {fn} in {cls}")

        if hasattr(fn, _method_name):
            logger.warning(
                f"Method {fn.__name__} already has a {cls.__name__} method. Overwriting..."
            )

        setattr(cls, _method_name, fn)

    if cls is None:
        return _inner_wrapper
    return _inner_wrapper(cls)


def register_from_pretrained(load_method: str | Callable, cls=None):
    _register_fn('from_pretrained', load_method, cls)


def register_save_pretrained(save_method: str | Callable, cls=None):
    _register_fn('save_pretrained', save_method, cls)


# class ConvertToJsonMixin(abc.ABC):

#     @abc.abstractmethod
#     def to_json(self):
#         """Convert the model to a JSON object."""
#         raise NotImplementedError("This method should be implemented in the subclass.")

#     @abc.abstractmethod
#     @classmethod
#     def from_json(cls, json_obj: dict):
#         raise NotImplementedError("This method should be implemented in the subclass.")


def is_json_serializable(obj):
    # if hasattr(obj, 'to_json') and hasattr(obj, 'from_json'):
    #     return True
    try:
        json.dumps(obj)
        return True
    except (TypeError, OverflowError):
        return False


class AutoSaver:

    config_name = "auto_saver_config.json"
    _auto_save_name = "_auto_save_modules"
    _pt_save_name = "_pt_save_modules"

    def __init__(self, **kwargs):
        self._constant = {}
        self._auto_save_modules = {}
        self._pt_save_modules = {}
        for name, value in kwargs.items():
            self.set(name, value)

    def set(self, key, value):
        # self._record[key] = value
        if hasattr(value, 'from_pretrained') and hasattr(value, 'save_pretrained'):
            self._auto_save_modules[key] = value
            return {self._auto_save_name: list(self._auto_save_modules.keys())}
        elif key in self._constant or is_json_serializable(value):
            self._constant[key] = value
            return {key: value}
        else:
            self._pt_save_modules[key] = value
            return {self._pt_save_name: list(self._pt_save_modules.keys())}

    def get(self, key):
        return ChainMap(
            self._constant, self._auto_save_modules, self._pt_save_modules
        ).get(key, None)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str, **kwargs):
        init_kwargs = cls.load(pretrained_model_name_or_path)
        return cls(**init_kwargs, **kwargs)

    @staticmethod
    def load(pretrained_model_name_or_path: str):
        config_path = os.path.join(pretrained_model_name_or_path, AutoSaver.config_name)

        config = load_json(config_path)

        _pt_save_module = {
            key: torch.load(os.path.join(pretrained_model_name_or_path, f'{key}.pt'))
            for key in config.pop(AutoSaver._pt_save_name, [])
            if not key.startswith('_')
        }

        _auto_save_module = {}
        for name, cls_name in config.pop(AutoSaver._auto_save_name, {}).items():
            if name.startswith('_'):
                continue
            builder = get_obj_from_str(cls_name)
            # try:
            # print(builder, name)
            module = builder.from_pretrained(
                os.path.join(pretrained_model_name_or_path, name)
            )

            _auto_save_module[name] = module

        config = {k: v for k, v in config.items() if not k.startswith("_")}

        init_kwargs = {**config, **_pt_save_module, **_auto_save_module}

        return init_kwargs

    def save_pretrained(self, path: str):
        # print(self._record)
        config_path = os.path.join(path, self.config_name)

        save_constant = copy.deepcopy(self._constant)
        save_constant[self._pt_save_name] = list(self._pt_save_modules.keys())

        save_constant[self._auto_save_name] = {}
        for name, value in self._auto_save_modules.items():
            save_constant[self._auto_save_name][
                name
            ] = f'{value.__class__.__module__}.{value.__class__.__name__}'
            name = name.lstrip('_')
            value.save_pretrained(os.path.join(path, name))

        safe_save_as_json(save_constant, config_path)

        for name, value in self._pt_save_modules.items():
            torch.save(value, os.path.join(path, f'{name}.pt'))

    @property
    def config(self):
        _config = copy.deepcopy(self._constant)
        _config[self._pt_save_name] = list(self._pt_save_modules.keys())
        _config[self._auto_save_name] = list(self._auto_save_modules.keys())
        return _config


def _register_to_autosave_init(init, ignore_list: List[str] = None):

    if ignore_list is None:
        ignore_list = []

    @functools.wraps(init)
    def inner_init(self, *args, **kwargs):

        # Ignore private kwargs in the init.
        init_kwargs = {k: v for k, v in kwargs.items() if not k.startswith("_")}
        config_init_kwargs = {k: v for k, v in kwargs.items() if k.startswith("_")}

        # Get positional arguments aligned with kwargs
        new_kwargs = {}
        signature = inspect.signature(init)
        parameters = {
            name: p.default
            for i, (name, p) in enumerate(signature.parameters.items())
            if i > 0
        }
        for arg, name in zip(args, parameters.keys()):
            new_kwargs[name] = arg

        # Then add all kwargs
        new_kwargs.update(
            {
                k: init_kwargs.get(k, default)
                for k, default in parameters.items()
                if k not in new_kwargs
            }
        )

        new_kwargs = {**config_init_kwargs, **new_kwargs}
        init(self, *args, **init_kwargs)

        self._phoenixcat_autosave_kwargs = {
            k: v for k, v in new_kwargs.items() if k not in ignore_list
        }

    return inner_init


def auto_register_save_load(
    cls,
    ignore_list: List[str] = None,
):
    cls.__init__ = _register_to_autosave_init(cls.__init__, ignore_list)

    @classmethod
    def from_pretrained(cls, pretrain_name_or_path, **kwargs):
        save_kwargs = AutoSaver.load(pretrain_name_or_path)
        # save_kwargs = {**save_kwargs, **kwargs}
        # init_kwargs, other_kwargs = split_init_other_parameters(cls, save_kwargs)
        init_kwargs = save_kwargs
        residule_kwargs = {}
        for key, value in kwargs.items():
            if key in init_kwargs:
                init_kwargs[key] = value
            else:
                residule_kwargs[key] = value
        residule_init_kwargs, residule_other_kwargs = split_init_other_parameters(
            cls, residule_kwargs
        )
        self = cls(**init_kwargs, **residule_init_kwargs)
        for key, value in residule_other_kwargs.items():
            setattr(self, key, value)
        return self

    def save_pretrained(self, save_directory):
        os.makedirs(save_directory, exist_ok=True)
        saver = AutoSaver()
        for name, value in self._phoenixcat_autosave_kwargs.items():
            saver.set(name, value)

        saver.save_pretrained(save_directory)

    cls.from_pretrained = from_pretrained
    cls.save_pretrained = save_pretrained

    return cls



from .path import get_safe_save_path
from .save import (
    safe_save_as_json,
    safe_save_as_yaml,
    safe_save_csv,
    safe_save_torchobj,
    safe_save_as_pickle,
)
from .walk import walk_extension_files, walk_images
from .load import load_csv, load_json, load_yaml, load_torchobj, load_pickle
from .manager import (
    CacheManager,
    FolderManager,
    DualFolderManager,
    RecordManager,
    set_record_manager_path,
    record_manager,
)



# Copyright 2024 Hongyao Yu.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import torch
from torch import nn
from torch.nn.parallel import DataParallel, DistributedDataParallel


def traverse_module(module: nn.Module, fn: Callable, call_middle=False):
    """Use DFS to traverse the module and visit submodules by function `fn`.

    Args:
        module (nn.Module): the module to be traversed
        fn (Callable): visit function
        call_middle (bool, optional): If true, it will visit both intermediate nodes and leaf nodes, else, it will only visit leaf nodes. Defaults to False.
    """

    children = list(module.children())
    if len(children) == 0:
        fn(module)
    else:
        if call_middle:
            fn(module)
        for child in children:
            traverse_module(child, fn)


def _traverse_name_module_impl(module_tuple: list, fn: Callable, call_middle=False):
    name, module = module_tuple
    children = list(module.named_children())
    if len(children) == 0:
        fn(module_tuple)
    else:
        if call_middle:
            fn(module_tuple)
        for child in children:
            _traverse_name_module_impl(child, fn)


def traverse_name_module(module: nn.Module, fn: Callable, call_middle=False):
    """Use DFS to traverse the module and visit submodules by function `fn`.

    Args:
        module (nn.Module): the module to be traversed
        fn (Callable): visit function
        call_middle (bool, optional): If true, it will visit both intermediate nodes and leaf nodes, else, it will only visit leaf nodes. Defaults to False.
    """
    children = list(module.named_children())
    for child in children:
        _traverse_name_module_impl(child, fn, call_middle=call_middle)


def freeze(net):
    for p in net.parameters():
        p.requires_grad_(False)


def unfreeze(net):
    for p in net.parameters():
        p.requires_grad_(True)


def unwrapped_parallel_module(module):

    if isinstance(module, (DataParallel, DistributedDataParallel)):
        return module.module
    return module



from typing import Union, Optional

import torch
import numpy as np


class ShapeException(Exception):
    pass


def check_shape(
    tensor: torch.Tensor | np.ndarray,
    expect_shape: Union[list[Optional[int]], list[Optional[int]]],
    raise_exception=True,
) -> bool:
    """Check if the shape of the tensor matches expectations.

    Args:
        tensor (torch.Tensor): The tensor to check.
        expect_shape (Union[list[Optional[int]], list[Optional[int]]]): The expected shape.
        raise_exception (bool, optional): Whether to raise an exception. Defaults to True.

    Returns:
        bool: The check result.
    """

    tensor_shape = tensor.shape

    if len(tensor_shape) < len(expect_shape):
        if raise_exception:
            raise ShapeException(
                f'expect ndim >= {len(expect_shape)}, but found {len(tensor_shape)}'
            )
        return False
    # torch.Size().
    tensor_shape_raw = tensor_shape
    tensor_shape = tensor_shape[-len(expect_shape) :]

    for i in range(len(expect_shape)):
        if expect_shape[i] is None:
            continue

        if expect_shape[i] != tensor_shape[i]:
            if raise_exception:
                for j in range(len(expect_shape)):
                    if expect_shape[i] is None:
                        expect_shape[i] = '*'
                expect_shape_str = ', '.join(expect_shape)
                tensor_shape_str = ', '.join(list(tensor_shape_raw))
                raise ShapeException(
                    f'expect shape [..., {expect_shape_str}], but found [{tensor_shape_str}]'
                )

            return False
    return True
